"""classes for storing signal data and plotting various spectra (IR, UV/vis, etc.)"""
import re

import numpy as np

from AaronTools import addlogger
from AaronTools.const import UNIT, PHYSICAL
from AaronTools.utils.utils import float_num


class Signal:
    """
    parent class for each signal in a spectrum
    """
    # attribute for the x position of this signal
    x_attr = None
    required_attrs = ()
    nested = None
    def __init__(self, x_var, **kwargs):
        for attr in self.required_attrs:
            setattr(self, attr, None)

        for arg in kwargs:
            setattr(self, arg, kwargs[arg])

        setattr(self, self.x_attr, x_var)


@addlogger
class Signals:
    """
    parent class for storing data for different signals in the
    spectrum and plotting a simulated spectrum
    """
    # label for x axis - should be set by child classes
    x_label = None
    
    LOG = None
    
    def __init__(self, data, style="gaussian", *args, **kwargs):
        self.data = []
        if isinstance(data[0], Signal):
            self.data = data
            return

        lines = False
        if isinstance(data, str):
            lines = data.splitlines()

        if lines and style == "gaussian":
            self.parse_gaussian_lines(lines, *args, **kwargs)
        elif lines and style == "orca":
            self.parse_orca_lines(lines, *args, **kwargs)
        elif lines and style == "psi4":
            self.parse_psi4_lines(lines, *args, **kwargs)
        elif lines and style == "qchem":
            self.parse_qchem_lines(lines, *args, **kwargs)
        else:
            raise NotImplementedError("cannot parse data for %s" % style)

    def parse_gaussian_lines(self, lines, *args, **kwargs):
        """parse data from Gaussian output files related to this spectrum"""
        raise NotImplementedError(
            "parse_gaussian_lines not implemented by %s" %
            self.__class__.__name__
        )

    def parse_orca_lines(self, lines, *args, **kwargs):
        """parse data from ORCA output files related to this spectrum"""
        raise NotImplementedError(
            "parse_orca_lines not implemented by %s" %
            self.__class__.__name__
        )

    def parse_psi4_lines(self, lines, *args, **kwargs):
        """parse data from Psi4 output files related to this spectrum"""
        raise NotImplementedError(
            "parse_psi4_lines not implemented by %s" %
            self.__class__.__name__
        )

    def parse_qchem_lines(self, lines, *args, **kwargs):
        """parse data from Q-Chem output files related to this spectrum"""
        raise NotImplementedError(
            "parse_qchem_lines not implemented by %s" %
            self.__class__.__name__
        )

    def filter_data(self, signal):
        """
        used to filter out some data from the spectrum (e.g.
        imaginary modes from an IR spec)
        return False if signal should not be in the spectrum
        """
        return True

    def get_spectrum_functions(
        self,
        fwhm=15.0,
        peak_type="pseudo-voigt",
        voigt_mixing=0.5,
        scalar_scale=0.0,
        linear_scale=0.0,
        quadratic_scale=0.0,
        intensity_attr="intensity",
        data_attr="data",
    ):
        """
        returns a list of functions that can be evaluated to
        produce a spectrum
        fwhm - full width at half max of each peak
        peak_type - gaussian, lorentzian, pseudo-voigt, or delta
        voigt_mixing - ratio of pseudo-voigt that is gaussian
        scalar_scale - shift x data
        linear_scale - scale x data
        quadratic_scale - scale x data
            x' = (1 - linear_scale * x - quadratic_scale * x^2 - scalar_scale)
        intensity_attr - attribute of Signal used for the intensity
            of that signal
        data_attr - attribute of self for the list of Signal()
        """
        data = getattr(self, data_attr)
        x_attr = data[0].x_attr
        
        
        # scale x positions
        if not data[0].nested:
            x_positions = np.array(
                [getattr(d, x_attr) for d in data if self.filter_data(d)]
            )
    
            intensities = [
                getattr(d, intensity_attr) for d in data if self.filter_data(d)
            ]
        else:
            x_positions = []
            intensities = []
            x_positions.extend(
                [getattr(d, x_attr) for d in data if self.filter_data(d)]
            )
            intensities.extend(
                [getattr(d, intensity_attr) for d in data if self.filter_data(d)]
            )
            for nest in data[0].nested:
                for d in data:
                    nest_attr = getattr(d, nest)
                    if isinstance(nest_attr, dict):
                        for value in nest_attr.values():
                            if hasattr(value, "__iter__"):
                                for item in value:
                                    x_positions.append(getattr(item, x_attr))
                                    intensities.append(getattr(item, intensity_attr))
                            else:
                                x_positions.append(getattr(value, x_attr))
                                intensities.append(getattr(value, intensity_attr))
                    elif hasattr(nest_attr, "__iter__"):
                        for item in nest_attr:
                            x_positions.append(getattr(item, x_attr))
                            intensities.append(getattr(item, intensity_attr))
                    else:
                        x_positions.append(getattr(nest_attr, x_attr))
                        intensities.append(getattr(nest_attr, intensity_attr))
                
            x_positions = np.array(x_positions)

        x_positions -= (
            linear_scale * x_positions + quadratic_scale * x_positions ** 2
        )
        x_positions += scalar_scale

        e_factor = -4 * np.log(2) / fwhm ** 2
        sigma = fwhm / (2 * np.sqrt(2 * np.log(2)))

        functions = []

        for x_pos, intensity in zip(x_positions, intensities):
            if intensity is not None:
                if peak_type.lower() == "gaussian":
                    functions.append(
                        lambda x, x0=x_pos, inten=intensity: inten
                        * np.exp(e_factor * (x - x0) ** 2)
                        * fwhm / (2 * np.sqrt(2 * np.log(2)))
                    )

                elif peak_type.lower() == "lorentzian":
                    functions.append(
                        lambda x, x0=x_pos, inten=intensity: inten
                        * (
                            0.5 * fwhm
                            / (np.pi * ((x - x0) ** 2 + (0.5 * fwhm) ** 2))
                        )
                    )

                elif peak_type.lower() == "pseudo-voigt":
                    functions.append(
                        lambda x, x0=x_pos, inten=intensity: inten
                        * (
                            (1 - voigt_mixing)
                            * (
                                (0.5 * fwhm) ** 2
                                / (((x - x0) ** 2 + (0.5 * fwhm) ** 2))
                            )
                            + voigt_mixing
                            * np.exp(-(x - x0) ** 2 / (2 * sigma ** 2))
                        )
                    )

                elif peak_type.lower() == "delta":
                    functions.append(
                        lambda x, x0=x_pos, inten=intensity: inten
                        * int(x == x0)
                    )

        return functions, x_positions, intensities

    @staticmethod
    def get_plot_data(
        functions,
        signal_centers,
        point_spacing=None,
        transmittance=False,
        peak_type="pseudo-voigt",
        normalize=True,
        fwhm=15.0,
        change_x_unit_func=None,
        show_functions=None,
    ):
        """
        returns arrays of x_values, y_values for a spectrum
        point_spacing - spacing between points, default is higher resolution around
                        each peak (i.e. not uniform)
                        this is pointless if peak_type == delta
        fwhm - full width at half max
        transmittance - if true, take 10^(2 - y_values) before returning
            to get transmittance as a %
        peak_type - pseudo-voigt, gaussian, lorentzian, or delta
        voigt_mixing - fraction of pseudo-voigt that is gaussian
        linear_scale - subtract linear_scale * frequency off each mode
        quadratic_scale - subtract quadratic_scale * frequency^2 off each mode
        """
        other_y_list = []
        if peak_type.lower() != "delta":
            if point_spacing is not None and peak_type.lower():
                x_values = []
                x = -point_spacing
                stop = max(signal_centers)
                if peak_type.lower() != "delta":
                    stop += 5 * fwhm
                while x < stop:
                    x += point_spacing
                    x_values.append(x)
            
                x_values = np.array(x_values)
            
            else:
                x_values = np.linspace(
                    0,
                    max(signal_centers) - 10 * fwhm,
                    num=100
                ).tolist()
            
                for freq in signal_centers:
                    x_values.extend(
                        np.linspace(
                            max(freq - (7.5 * fwhm), 0),
                            freq + (7.5 * fwhm),
                            num=75,
                        ).tolist()
                    )
                    x_values.append(freq)

                if not point_spacing:
                    x_values = np.array(list(set(x_values)))
                    x_values.sort()

            y_values = np.sum([f(x_values) for f in functions], axis=0)
            
            if show_functions:
                for (ndx1, ndx2) in show_functions:
                    other_y_list.append(
                        np.sum(
                            [f(x_values) for f in functions[ndx1: ndx2]],
                            axis=0,
                        )
                    )

        else:
            x_values = []
            y_values = []

            for freq, func in zip(signal_centers, functions):
                y_values.append(func(freq))
                x_values.append(freq)

            y_values = np.array(y_values)

        if len(y_values) == 0:
            Signals.LOG.warning("nothing to plot")
            return None

        if normalize or transmittance:
            max_val = abs(max(y_values.max(), y_values.min(), key=abs))
            y_values /= max_val
            for y_vals in other_y_list:
                y_vals /= max_val

        if transmittance:
            y_values = np.array([10 ** (2 - y) for y in y_values])
            for i in range(0, len(other_y_list)):
                other_y_list[i] = np.array(
                    [10 ** (2 - y) for y in other_y_list[i]]
                )

        if change_x_unit_func:
            x_values, ndx = change_x_unit_func(x_values)
            y_values = y_values[ndx]
            for i in range(0, len(other_y_list)):
                other_y_list[i] = other_y_list[i][ndx]

        return x_values, y_values, other_y_list

    @classmethod
    def plot_spectrum(
        cls,
        figure,
        x_values,
        y_values,
        other_y_values=None,
        other_y_style=None,
        centers=None,
        widths=None,
        exp_data=None,
        reverse_x=None,
        y_label=None,
        plot_type="transmittance",
        x_label=r"wavenumber (cm$^{-1}$)",
        peak_type="pseudo-voigt",
        rotate_x_ticks=False,
    ):
        """
        plot the x_data and y_data on figure (matplotlib figure)
        this is intended for IR spectra

        centers - array-like of float, plot is split into sections centered
                  on the frequency specified by centers
                  default is to not split into sections
        widths - array-like of float, defines the width of each section
        exp_data - other data to plot
                   should be a list of (x_data, y_data, color)
        reverse_x - if True, 0 cm^-1 will be on the right
        """

        if not centers:
            # if no centers were specified, pretend they were so we
            # can do everything the same way
            axes = [figure.subplots(nrows=1, ncols=1)]
            y_nonzero = np.nonzero(y_values)[0]
            x_values = np.array(x_values)
            widths = [max(x_values[y_nonzero])]
            centers = [max(x_values[y_nonzero]) / 2]
        else:
            n_sections = len(centers)
            figure.subplots_adjust(wspace=0.05)
            # sort the sections so we don't jump around
            widths = [
                x
                for _, x in sorted(
                    zip(centers, widths),
                    key=lambda p: p[0],
                    reverse=reverse_x,
                )
            ]
            centers = sorted(centers, reverse=reverse_x)

            axes = figure.subplots(
                nrows=1,
                ncols=n_sections,
                sharey=True,
                gridspec_kw={"width_ratios": widths},
            )
            if not hasattr(axes, "__iter__"):
                # only one section was specified (e.g. zooming in on a peak)
                # make sure axes is iterable
                axes = [axes]

        for i, ax in enumerate(axes):
            if i == 0:
                ax.set_ylabel(y_label)

                # need to split plot into sections
                # put a / on the border at the top and bottom borders
                # of the plot
                if len(axes) > 1:
                    ax.spines["right"].set_visible(False)
                    ax.tick_params(labelright=False, right=False)
                    ax.plot(
                        [1, 1],
                        [0, 1],
                        marker=((-1, -1), (1, 1)),
                        markersize=5,
                        linestyle="none",
                        color="k",
                        mec="k",
                        mew=1,
                        clip_on=False,
                        transform=ax.transAxes,
                    )

            elif i == len(axes) - 1 and len(axes) > 1:
                # last section needs a set of / too, but on the left side
                ax.spines["left"].set_visible(False)
                ax.tick_params(labelleft=False, left=False)
                ax.plot(
                    [0, 0],
                    [0, 1],
                    marker=((-1, -1), (1, 1)),
                    markersize=5,
                    linestyle="none",
                    color="k",
                    mec="k",
                    mew=1,
                    clip_on=False,
                    transform=ax.transAxes,
                )

            elif len(axes) > 1:
                # middle sections need two sets of /
                ax.spines["right"].set_visible(False)
                ax.spines["left"].set_visible(False)
                ax.tick_params(
                    labelleft=False, labelright=False, left=False, right=False
                )
                ax.plot(
                    [0, 0],
                    [0, 1],
                    marker=((-1, -1), (1, 1)),
                    markersize=5,
                    linestyle="none",
                    label="Silence Between Two Subplots",
                    color="k",
                    mec="k",
                    mew=1,
                    clip_on=False,
                    transform=ax.transAxes,
                )
                ax.plot(
                    [1, 1],
                    [0, 1],
                    marker=((-1, -1), (1, 1)),
                    markersize=5,
                    label="Silence Between Two Subplots",
                    linestyle="none",
                    color="k",
                    mec="k",
                    mew=1,
                    clip_on=False,
                    transform=ax.transAxes,
                )

            if peak_type.lower() != "delta":
                ax.plot(
                    x_values,
                    y_values,
                    color="k",
                    linewidth=1,
                    label="computed",
                )
                
                if other_y_values:
                    for y_vals, style in zip(other_y_values, other_y_style):
                        ax.plot(
                            x_values,
                            y_vals,
                            color=style[0],
                            linestyle=style[1],
                            linewidth=1,
                            label=style[2],
                            zorder=-1,
                        )

            else:
                if plot_type.lower() == "transmittance":
                    ax.vlines(
                        x_values,
                        y_values,
                        [100 for y in y_values],
                        linewidth=1,
                        colors=["k" for x in x_values],
                        label="computed",
                    )
                    ax.hlines(
                        100,
                        0,
                        max(4000, *x_values),
                        linewidth=1,
                        colors=["k" for y in y_values],
                        label="computed",
                    )
                    
                    if other_y_values:
                        for y_vals, style in zip(other_y_values, other_y_style):
                            ax.vlines(
                                x_values,
                                y_vals,
                                [100 for y in y_vals],
                                colors=[style[0] for x in x_values],
                                linestyles=style[1],
                                linewidth=1,
                                label=style[2],
                                zorder=-1,
                            )
                else:
                    ax.vlines(
                        x_values,
                        [0 for y in y_values],
                        y_values,
                        linewidth=1,
                        colors=["k" for x in x_values],
                        label="computed",
                    )
                    ax.hlines(
                        0,
                        0,
                        max(4000, *x_values),
                        linewidth=1,
                        colors=["k" for y in y_values],
                        label="computed",
                    )
                    
                    if other_y_values:
                        for y_vals, style in zip(other_y_values, other_y_style):
                            ax.vlines(
                                x_values,
                                [0 for y in y_vals],
                                y_vals,
                                colors=[style[0] for x in x_values],
                                linestyles=style[1],
                                linewidth=1,
                                label=style[2],
                                zorder=-1,
                            )

            if exp_data:
                for x, y, color in exp_data:
                    ax.plot(
                        x,
                        y,
                        color=color,
                        zorder=1,
                        linewidth=1,
                        label="observed",
                    )

            center = centers[i]
            width = widths[i]
            high = center + width / 2
            low = center - width / 2
            if reverse_x:
                ax.set_xlim(high, low)
            else:
                ax.set_xlim(low, high)

        # b/c we're doing things in sections, we can't add an x-axis label
        # well we could, but which section would be put it one?
        # it wouldn't be centered
        # so instead the x-axis label is this
        figure.text(
            0.5, 0.0, x_label, ha="center", va="bottom"
        )
        if rotate_x_ticks:
            figure.autofmt_xdate(rotation=-45, ha="center")

    @classmethod
    def get_mixed_signals(
        cls,
        signal_groups,
        weights,
        fractions=None,
        data_attr="data",
        **kwargs,
    ):
        """
        get signals for a mixture of components or conformers
        signal_groups - list of Signals() instances or list of lists of Signals()
            a list of Signals() is a group of conformers
            a list of lists of Signals() are the different components
        weights - weights for each conformer, organized according to signal_groups
        fractions - fraction of each component in the mixture
            default: all components have equal fractions
        data_attr - attribute of Signals() for data
        **kwargs - passed to cls.__init__, along with a new list of data
        """
        if not hasattr(signal_groups[0], "__iter__"):
            signal_groups = [signal_groups]
        
        if not hasattr(weights[0], "__iter__"):
            weights = [weights]
        
        if fractions is None:
            fractions = np.ones(len(signal_groups))
        
        new_data = []
        for group, weighting, fraction in zip(signal_groups, weights, fractions):
            for signals, weight in zip(group, weighting):
                data = getattr(signals, data_attr)
                for d in data:
                    x_val = getattr(d, d.x_attr)
                    vals = d.__dict__
                    data_cls = d.__class__
                    new_vals = dict()
                    for key, item in vals.items():
                        if isinstance(item, float):
                            new_vals[key] = fraction * weight * item
                        else:
                            new_vals[key] = item
                    
                    if d.nested:
                        if not isinstance(d.nested, str):
                            for attr in d.nested:
                                nest = getattr(d, attr)
                                nest_vals = dict()
                                if isinstance(nest, dict):
                                    for k, items in nest.items():
                                        for i, item in enumerate(items):
                                            nest_x_val = getattr(item, item.x_attr)
                                            vals = item.__dict__
                                            nest_cls = item.__class__
                                            for k2, j in vals.items():
                                                if isinstance(item, float):
                                                    nest_vals[k2] = fraction * weight * j
                                                else:
                                                    nest_vals[k2] = j
                                            new_vals[attr][k][i] = nest_cls(
                                                nest_x_val, **nest_vals
                                            )
                                elif hasattr(nest, "__iter__"):
                                    for i, item in enumerate(nest):
                                        nest_x_val = getattr(item, item.x_attr)
                                        vals = item.__dict__
                                        nest_cls = item.__class__
                                        for k, j in vals.items():
                                            if isinstance(item, float):
                                                nest_vals[k] = fraction * weight * j
                                            else:
                                                nest_vals[k] = j
                                        new_vals[attr][i] = nest_cls(
                                            nest_x_val, **nest_vals
                                        )
                                else:
                                    nest_x_val = getattr(nest, nest.x_attr)
                                    vals = nest.__dict__
                                    nest_cls = nest.__class__
                                    for k, j in vals.items():
                                        if isinstance(nest, float):
                                            nest_vals[k] = fraction * weight * j
                                        else:
                                            nest_vals[k] = j
                                    new_vals[attr][k] = nest_cls(
                                        nest_x_val, **nest_vals
                                    )
                    
                    new_data.append(data_cls(x_val, **new_vals))

        return cls(new_data, **kwargs)


class HarmonicVibration(Signal):
    x_attr = "frequency"
    required_attrs = (
        "intensity", "vector", "symmetry", "rotation", "raman_activity", "forcek",
    )


class AnharmonicVibration(Signal):
    x_attr = "frequency"
    required_attrs = (
        "intensity", "harmonic", "overtones", "combinations",
        "rotation", "raman_activity",
    )
    nested = ("overtones", "combinations")

    @property
    def harmonic_frequency(self):
        return self.harmonic.frequency

    @property
    def delta_anh(self):
        return self.frequency - self.harmonic.frequency


class Frequency(Signals):
    """for spectra in the IR/NIR region based on vibrational modes"""
    
    def __init__(self, *args, harmonic=True, hpmodes=None, **kwargs):
        super().__init__(*args, harmonic=harmonic, hpmodes=hpmodes, **kwargs)
        self.anharm_data = None
        self.imaginary_frequencies = None
        self.real_frequencies = None
        self.lowest_frequency = None
        self.by_frequency = {}
        self.is_TS = None
        self.sort_frequencies()

    def parse_gaussian_lines(
        self, lines, *args, hpmodes=None, harmonic=True, **kwargs
    ):
        if harmonic:
            return self._parse_harmonic_gaussian(lines, hpmodes=hpmodes)
        return self._parse_anharmonic_gaussian(lines)

    def _parse_harmonic_gaussian(self, lines, hpmodes):
        if hpmodes is None:
            raise TypeError(
                "hpmodes argument required when data is a string"
            )
        num_head = 0
        for line in lines:
            if "Harmonic frequencies" in line:
                num_head += 1
        if hpmodes and num_head != 2:
            raise RuntimeError("Log file damaged, cannot get frequencies")

        num_head = 0
        idx = -1
        modes = []
        for k, line in enumerate(lines):
            if "Harmonic frequencies" in line:
                num_head += 1
                if hpmodes and num_head == 2:
                    # if hpmodes, want just the first set of freqs
                    break
                continue
            if "Frequencies" in line and (
                (hpmodes and "---" in line) or ("--" in line and not hpmodes)
            ):
                for i, symm in zip(
                    float_num.findall(line), lines[k - 1].split()
                ):
                    self.data += [HarmonicVibration(float(i), symmetry=symm)]
                    modes += [[]]
                    idx += 1
                continue

            if ("Force constants" in line and "---" in line and hpmodes) or (
                "Frc consts" in line and "--" in line and not hpmodes
            ):
                force_constants = float_num.findall(line)
                for i in range(-len(force_constants), 0, 1):
                    self.data[i].forcek = float(force_constants[i])
                continue

            if ("Reduced masses" in line and "---" in line and hpmodes) or (
                "Red. masses" in line and "--" in line and not hpmodes
            ):
                red_masses = float_num.findall(line)
                for i in range(-len(red_masses), 0, 1):
                    self.data[i].red_mass = float(red_masses[i])
                continue

            if ("Rot. strength" in line and "---" in line and hpmodes) or (
                "Rot. str." in line and "--" in line and not hpmodes
            ):
                roational_strength = float_num.findall(line)
                for i in range(-len(roational_strength), 0, 1):
                    self.data[i].rotation = float(roational_strength[i])
                continue

            if ("Raman Activities" in line and "---" in line and hpmodes) or (
                "Raman Activ" in line and "--" in line and not hpmodes
            ):
                roational_strength = float_num.findall(line)
                for i in range(-len(roational_strength), 0, 1):
                    self.data[i].raman_activity = float(roational_strength[i])
                continue

            if "IR Inten" in line and (
                (hpmodes and "---" in line) or (not hpmodes and "--" in line)
            ):
                intensities = float_num.findall(line)
                for i in range(-len(force_constants), 0, 1):
                    self.data[i].intensity = float(intensities[i])
                continue

            if hpmodes:
                match = re.search(
                    r"^\s+\d+\s+\d+\s+\d+(\s+[+-]?\d+\.\d+)+$", line
                )
                if match is None:
                    continue
                values = float_num.findall(line)
                coord = int(values[0]) - 1
                atom = int(values[1]) - 1
                moves = values[3:]
                for i, m in enumerate(moves):
                    tmp = len(moves) - i
                    mode = modes[-tmp]
                    try:
                        vector = mode[atom]
                    except IndexError:
                        vector = [0, 0, 0]
                        modes[-tmp] += [[]]
                    vector[coord] = m
                    modes[-tmp][atom] = vector
            else:
                match = re.search(r"^\s+\d+\s+\d+(\s+[+-]?\d+\.\d+)+$", line)
                if match is None:
                    continue
                values = float_num.findall(line)
                atom = int(values[0]) - 1
                moves = np.array(values[2:], dtype=np.float)
                n_moves = len(moves) // 3
                for i in range(-n_moves, 0):
                    modes[i].append(
                        moves[3 * n_moves + 3 * i : 4 * n_moves + 3 * i]
                    )

        for mode, data in zip(modes, self.data):
            data.vector = np.array(mode, dtype=np.float64)

    def _parse_anharmonic_gaussian(self, lines):
        reading_combinations = False
        reading_overtones = False
        reading_fundamentals = False

        combinations = []
        overtones = []
        fundamentals = []

        mode_re = re.compile(r"(\d+)\((\d+)\)")

        for line in lines:
            if "---" in line or "Mode" in line or not line.strip():
                continue
            if "Fundamental Bands" in line:
                reading_fundamentals = True
                continue
            if "Overtones" in line:
                reading_overtones = True
                continue
            if "Combination Bands" in line:
                reading_combinations = True
                continue

            if reading_combinations:
                info = line.split()
                mode1 = mode_re.search(info[0])
                mode2 = mode_re.search(info[1])
                ndx_1 = int(mode1.group(1))
                exp_1 = int(mode1.group(2))
                ndx_2 = int(mode2.group(1))
                exp_2 = int(mode2.group(2))
                harm_freq = float(info[2])
                anharm_freq = float(info[3])
                anharm_inten = float(info[4])
                harm_inten = 0
                combinations.append(
                    (
                        ndx_1,
                        ndx_2,
                        exp_1,
                        exp_2,
                        anharm_freq,
                        anharm_inten,
                        harm_freq,
                        harm_inten,
                    )
                )
            elif reading_overtones:
                info = line.split()
                mode = mode_re.search(info[0])
                ndx = int(mode.group(1))
                exp = int(mode.group(2))
                harm_freq = float(info[1])
                anharm_freq = float(info[2])
                anharm_inten = float(info[3])
                harm_inten = 0
                overtones.append(
                    (
                        ndx,
                        exp,
                        anharm_freq,
                        anharm_inten,
                        harm_freq,
                        harm_inten,
                    )
                )
            elif reading_fundamentals:
                info = line.split()
                harm_freq = float(info[1])
                anharm_freq = float(info[2])
                anharm_inten = float(info[4])
                harm_inten = float(info[3])
                fundamentals.append(
                    (anharm_freq, anharm_inten, harm_freq, harm_inten)
                )

        self.anharm_data = []
        for i, mode in enumerate(
            sorted(fundamentals, key=lambda pair: pair[2])
        ):
            self.anharm_data.append(
                AnharmonicVibration(mode[0], intensity=mode[1], harmonic=self.data[i])
            )
            self.anharm_data[-1].overtones = []
            self.anharm_data[-1].combinations = dict()

        for overtone in overtones:
            ndx = len(fundamentals) - overtone[0]
            data = self.anharm_data[ndx]
            harm_data = HarmonicVibration(overtone[4], intensity=overtone[5])
            data.overtones.append(
                AnharmonicVibration(
                    overtone[2], intensity=overtone[3], harmonic=harm_data
                )
            )
        for combo in combinations:
            ndx1 = len(fundamentals) - combo[0]
            ndx2 = len(fundamentals) - combo[1]
            data = self.anharm_data[ndx1]
            harm_data = HarmonicVibration(combo[6], intensity=combo[7])
            data.combinations[ndx2] = [
                AnharmonicVibration(combo[4], intensity=combo[5], harmonic=harm_data)
            ]
    
    def parse_qchem_lines(self, lines, *args, **kwargs):
        num_head = 0
        modes = []
        for k, line in enumerate(lines):
            if "Frequency:" in line:
                ndx = 0
                for i in float_num.findall(line):
                    self.data += [HarmonicVibration(float(i))]
                    modes += [[]]
                continue

            if "Force Cnst:" in line:
                force_constants = float_num.findall(line)
                for i in range(-len(force_constants), 0, 1):
                    self.data[i].forcek = float(force_constants[i])
                continue

            if "Red. Mass:" in line:
                red_masses = float_num.findall(line)
                for i in range(-len(red_masses), 0, 1):
                    self.data[i].red_mass = float(red_masses[i])
                continue

            if "IR Intens:" in line:
                intensities = float_num.findall(line)
                for i in range(-len(force_constants), 0, 1):
                    self.data[i].intensity = float(intensities[i])
                continue

            if "Raman Intens:" in line:
                intensities = float_num.findall(line)
                for i in range(-len(force_constants), 0, 1):
                    self.data[i].raman_activity = float(intensities[i])
                continue

            match = re.search(r"^\s?[A-Z][a-z]?\s+(\s+[+-]?\d+\.\d+)+$", line)
            if match is None:
                continue
            ndx += 1
            values = float_num.findall(line)
            moves = np.array(values, dtype=np.float)
            n_moves = len(moves) // 3
            for i in range(-n_moves, 0):
                modes[i].append(
                    moves[3 * n_moves + 3 * i : 4 * n_moves + 3 * i]
                )

        for mode, data in zip(modes, self.data):
            data.vector = np.array(mode, dtype=np.float64)

    def parse_orca_lines(self, lines, *args, **kwargs):
        """parse lines of orca output related to frequency
        hpmodes is not currently used"""
        # vibrational frequencies appear as a list, one per line
        # block column 0 is the index of the mode
        # block column 1 is the frequency in 1/cm
        # skip line one b/c its just "VIBRATIONAL FREQUENCIES" with the way we got the lines
        for n, line in enumerate(lines[1:]):
            if line == "NORMAL MODES":
                break

            freq = line.split()[1]
            self.data += [HarmonicVibration(float(freq))]

        # all 3N modes are printed with six modes in each block
        # each column corresponds to one mode
        # the rows of the columns are x_1, y_1, z_1, x_2, y_2, z_2, ...
        displacements = np.zeros((len(self.data), len(self.data)))
        carryover = 0
        start = 0
        stop = 6
        for i, line in enumerate(lines[n + 2 :]):
            if "IR SPECTRUM" in line:
                break

            if i % (len(self.data) + 1) == 0:
                carryover = i // (len(self.data) + 1)
                start = 6 * carryover
                stop = start + 6
                continue

            ndx = (i % (len(self.data) + 1)) - 1
            mode_info = line.split()[1:]

            displacements[ndx][start:stop] = [float(x) for x in mode_info]

        # reshape columns into Nx3 arrays
        for k, data in enumerate(self.data):
            data.vector = np.reshape(
                displacements[:, k], (len(self.data) // 3, 3)
            )

        # purge rotational and translational modes
        n_data = len(self.data)
        k = 0
        while k < n_data:
            if self.data[k].frequency == 0:
                del self.data[k]
                n_data -= 1
            else:
                k += 1

        for k, line in enumerate(lines):
            if line.strip() == "IR SPECTRUM":
                order = lines[k + 1].split()
                if "Int" in order:
                    ndx = order.index("Int")
                else:
                    ndx = order.index("T**2") - 1
                intensity_start = k + 2

        # IR intensities are only printed for vibrational
        # the first column is the index of the mode
        # the second column is the frequency
        # the third is the intensity, which we read next
        t = 0
        for line in lines[intensity_start:]:
            if not re.match(r"\s*\d+:", line):
                continue
            ir_info = line.split()
            inten = float(ir_info[ndx])
            self.data[t].intensity = inten
            t += 1
            if t >= len(self.data):
                break

        for k, line in enumerate(lines):
            if line.strip() == "RAMAN SPECTRUM":
                t = 0
                for line in lines[k + 1:]:
                    if not re.match(r"\s*\d+:", line):
                        continue
                    ir_info = line.split()
                    inten = float(ir_info[2])
                    self.data[t].raman_activity = inten
                    t += 1
                    if t >= len(self.data):
                        break

    def parse_psi4_lines(self, lines, *args, **kwargs):
        """parse lines of psi4 output related to frequencies
        hpmodes is not used"""
        # normal mode info appears in blocks, with up to 3 modes per block
        # at the top is the index of the normal mode
        # next is the frequency in wavenumbers (cm^-1)
        # after a line of '-----' are the normal displacements
        read_displacement = False
        modes = []
        for n, line in enumerate(lines):
            if len(line.strip()) == 0:
                read_displacement = False
                for i, data in enumerate(self.data[-nmodes:]):
                    data.vector = np.array(modes[i])

            elif read_displacement:
                info = [float(x) for x in line.split()[2:]]
                for i, mode in enumerate(modes):
                    mode.append(info[3 * i : 3 * (i + 1)])

            elif line.strip().startswith("Vibration"):
                nmodes = len(line.split()) - 1

            elif line.strip().startswith("Freq"):
                freqs = [-1 * float(x.strip("i")) if x.endswith("i") else float(x) for x in line.split()[2:]]
                for freq in freqs:
                    self.data.append(HarmonicVibration(float(freq)))

            elif line.strip().startswith("Force const"):
                force_consts = [float(x) for x in line.split()[3:]]
                for i, data in enumerate(self.data[-nmodes:]):
                    data.forcek = force_consts[i]

            elif line.strip().startswith("Irrep"):
                # sometimes psi4 doesn't identify the irrep of a mode, so we can't
                # use line.split()
                symm = [
                    x.strip() if x.strip() else None
                    for x in [line[31:40], line[51:60], line[71:80]]
                ]
                for i, data in enumerate(self.data[-nmodes:]):
                    data.symmetry = symm[i]

            elif line.strip().startswith("----"):
                read_displacement = True
                modes = [[] for i in range(0, nmodes)]

    def sort_frequencies(self):
        self.imaginary_frequencies = []
        self.real_frequencies = []
        for data in self.data:
            freq = data.frequency
            if freq < 0:
                self.imaginary_frequencies += [freq]
            elif freq > 0:
                self.real_frequencies += [freq]
            self.by_frequency[freq] = {
                "intensity": data.intensity,
            }
            if hasattr(data, "vector"):
                # anharmonic data might not have a vector
                self.by_frequency[freq]["vector"] = data.vector
        if len(self.data) > 0:
            self.lowest_frequency = self.data[0].frequency
        else:
            self.lowest_frequency = None
        self.is_TS = True if len(self.imaginary_frequencies) == 1 else False

    def filter_data(self, signal):
        return signal.frequency > 0

    def plot_ir(
        self,
        figure,
        centers=None,
        widths=None,
        exp_data=None,
        plot_type="transmittance",
        peak_type="pseudo-voigt",
        reverse_x=True,
        y_label=None,
        point_spacing=None,
        normalize=True,
        fwhm=15.0,
        anharmonic=False,
        rotate_x_ticks=False,
        show_functions=None,
        **kwargs,
    ):
        """
        plot IR data on figure
        figure - matplotlib figure
        centers - array-like of float, plot is split into sections centered
                  on the frequency specified by centers
                  default is to not split into sections
        widths - array-like of float, defines the width of each section
        exp_data - other data to plot
                   should be a list of (x_data, y_data, color)
        reverse_x - if True, 0 cm^-1 will be on the right
        plot_type - see Frequency.get_plot_data
        peak_type - any value allowed by Frequency.get_plot_data
        kwargs - keywords for Frequency.get_spectrum_functions
        """

        if "intensity_attr" not in kwargs:
            intensity_attr = "intensity"
            if plot_type.lower() == "vcd":
                intensity_attr = "rotation"
            elif plot_type.lower() == "raman":
                intensity_attr = "raman_activity"
            elif plot_type.lower() == "absorbance":
                intensity_attr = "intensity"
            elif plot_type.lower() == "transmittance":
                intensity_attr = "intensity"
            else:
                self.LOG.warning("unrecognized plot type: %s\nDefaulting to absorbance" % plot_type)
            kwargs["intensity_attr"] = intensity_attr

        data_attr = "data"
        if anharmonic:
            data_attr = "anharm_data"

        functions, frequencies, intensities = self.get_spectrum_functions(
            peak_type=peak_type,
            fwhm=fwhm,
            data_attr=data_attr,
            **kwargs,
        )

        other_y_style = None
        ndx_list = None
        if show_functions is not None:
            ndx_list = [info[0] for info in show_functions]
            other_y_style = list(info[1:] for info in show_functions)

        data = self.get_plot_data(
            functions,
            frequencies,
            fwhm=fwhm,
            transmittance=plot_type.lower().startswith("transmittance"),
            peak_type=peak_type,
            point_spacing=point_spacing,
            normalize=normalize,
            show_functions=ndx_list,
        )
        if data is None:
            return

        x_values, y_values, other_y_values = data

        if y_label is None and plot_type.lower().startswith("transmittance"):
            y_label = "Transmittance (%)"
        elif y_label is None and plot_type.lower() == "absorbance":
            y_label = "Absorbance (arb.)"
        elif y_label is None and plot_type.lower() == "vcd":
            y_label = "ΔAbsorbance (arb.)"
        elif y_label is None and plot_type.lower() == "raman":
            y_label = "Activity (arb.)"

        self.plot_spectrum(
            figure,
            x_values,
            y_values,
            other_y_values=other_y_values,
            other_y_style=other_y_style,
            centers=centers,
            widths=widths,
            exp_data=exp_data,
            reverse_x=reverse_x,
            peak_type=peak_type,
            plot_type=plot_type,
            y_label=y_label,
            rotate_x_ticks=rotate_x_ticks,
        )


class ValenceExcitation(Signal):
    x_attr = "excitation_energy"
    required_attrs = (
        "rotatory_str_len", "rotatory_str_vel", "oscillator_str",
        "oscillator_str_vel", "symmetry", "multiplicity",
    )

    @property
    def dipole_str_len(self):
        return self.oscillator_str / self.excitation_energy

    @property
    def dipole_str_vel(self):
        return self.oscillator_str_vel / self.excitation_energy


class TransientExcitation(ValenceExcitation):
    x_attr = "excitation_energy"
    required_attrs = (
        "rotatory_str_len", "rotatory_str_vel", "oscillator_str",
        "oscillator_str_vel", "symmetry", "multiplicity",
    )
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class ValenceExcitations(Signals):
    def __init__(self, *args, **kwargs):
        self.transient_data = None
        super().__init__(*args, **kwargs)

    def parse_gaussian_lines(self, lines, *args, **kwargs):
        i = 0
        nrgs = []
        rotatory_str_len = []
        rotatory_str_vel = []
        oscillator_str = []
        oscillator_vel = []
        symmetry = []
        multiplicity = []
        while i < len(lines):
            if "Ground to excited state transition electric" in lines[i]:
                i += 2
                line = lines[i]
                while line and line.split()[0].isdigit():
                    oscillator_str.append(float(line.split()[-1]))
                    i += 1
                    line = lines[i]
            elif "Ground to excited state transition velocity" in lines[i]:
                i += 2
                line = lines[i]
                while line and line.split()[0].isdigit():
                    oscillator_vel.append(float(line.split()[-1]))
                    i += 1
                    line = lines[i]
            elif "R(length)" in lines[i]:
                i += 1
                line = lines[i]
                while line and line.split()[0].isdigit():
                    rotatory_str_len.append(float(line.split()[-1]))
                    i += 1
                    line = lines[i]
            elif "R(velocity)" in lines[i]:
                i += 1
                line = lines[i]
                while line and line.split()[0].isdigit():
                    rotatory_str_vel.append(float(line.split()[-2]))
                    i += 1
                    line = lines[i]
            elif re.search(r"Excited State\s*\d+:", lines[i]):
                excitation_data = re.search(
                    r"Excited State\s*\d+:\s*([\D]+)-([\S]+)\s+(\d+\.\d+)",
                    lines[i],
                )
                multiplicity.append(excitation_data.group(1))
                symmetry.append(excitation_data.group(2))
                nrgs.append(float(excitation_data.group(3)))
                i += 1
            else:
                i += 1

        for nrg, rot_len, rot_vel, osc_len, osc_vel, sym, mult in zip(
            nrgs, rotatory_str_len, rotatory_str_vel, oscillator_str,
            oscillator_vel, symmetry, multiplicity,
        ):
            self.data.append(
                ValenceExcitation(
                    nrg, rotatory_str_len=rot_len,
                    rotatory_str_vel=rot_vel, oscillator_str=osc_len,
                    oscillator_str_vel=osc_vel, symmetry=sym,
                    multiplicity=mult,
                )
            )

    def parse_orca_lines(self, lines, *args, **kwargs):
        i = 0
        nrgs = []
        corr = []
        rotatory_str_len = []
        rotatory_str_vel = []
        oscillator_str = []
        oscillator_vel = []
        multiplicity = []
        mult = "Singlet"
        
        transient_oscillator_str = []
        transient_oscillator_vel = []
        transient_rot_str = []
        transient_nrg = []
        
        while i < len(lines):
            line = lines[i]
            if "SINGLETS" in line:
                mult = "Singlet"
                i += 1
            elif "TRIPLETS" in line:
                mult = "Triplet"
                i += 1
            elif re.search("IROOT=.+?(\d+\.\d+)\seV", line):
                info = re.search("IROOT=.+?(\d+\.\d+)\seV", line)
                nrgs.append(float(info.group(1)))
                i += 1
            elif line.startswith("STATE"):
                info = re.search("STATE\s*\d+:\s*E=\s*\S+\s*au\s*(\d+\.\d+)", line)
                nrgs.append(float(info.group(1)))
                multiplicity.append(mult)
                i += 1
            elif "ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS" in line and "TRANSIENT" not in line:
                i += 5
                line = lines[i]
                while line.strip():
                    info = line.split()
                    oscillator_str.append(float(info[3]))
                    i += 1
                    line = lines[i]
            elif "ABSORPTION SPECTRUM VIA TRANSITION VELOCITY DIPOLE MOMENTS" in line and "TRANSIENT" not in line:
                i += 5
                line = lines[i]
                while line.strip():
                    info = line.split()
                    oscillator_vel.append(float(info[3]))
                    i += 1
                    line = lines[i]
            elif line.endswith("CD SPECTRUM") and "TRANSIENT" not in line:
                i += 5
                line = lines[i]
                while line.strip():
                    info = line.split()
                    rotatory_str_len.append(float(info[3]))
                    i += 1
                    line = lines[i]
            elif "CD SPECTRUM VIA TRANSITION VELOCITY DIPOLE MOMENTS" in line and "TRANSIENT" not in line:
                i += 5
                line = lines[i]
                while line.strip():
                    info = line.split()
                    rotatory_str_vel.append(float(info[3]))
                    i += 1
                    line = lines[i]
            elif "TRANSIENT ABSORPTION SPECTRUM VIA TRANSITION ELECTRIC DIPOLE MOMENTS" in line:
                i += 5
                line = lines[i]
                while line.strip():
                    info = line.split()
                    transient_oscillator_str.append(float(info[3]))
                    transient_nrg.append(self.nm_to_ev(float(info[2])))
                    i += 1
                    line = lines[i]
            elif "TRANSIENT ABSORPTION SPECTRUM VIA TRANSITION VELOCITY DIPOLE MOMENTS" in line:
                i += 5
                line = lines[i]
                while line.strip():
                    info = line.split()
                    transient_oscillator_vel.append(float(info[3]))
                    i += 1
                    line = lines[i]
            elif "TRANSIENT CD SPECTRUM" in line:
                i += 5
                line = lines[i]
                while line.strip():
                    info = line.split()
                    transient_rot_str.append(float(info[3]))
                    i += 1
                    line = lines[i]
            elif "CALCULATED SOLVENT SHIFTS" in line:
                i += 8
                line = lines[i]
                while line.strip():
                    info = line.split()
                    corr.append(float(info[-1]))
                    i += 1
                    line = lines[i]
                
            else:
                i += 1

        if corr:
            for i in range(0, len(nrgs)):
                nrgs[i] = corr[i]

        if not multiplicity:
            multiplicity = [None for x in nrgs]

        if not rotatory_str_vel:
            rotatory_str_vel = [None for x in rotatory_str_len]

        if not oscillator_vel:
            oscillator_vel = [None for x in oscillator_str]

        for nrg, rot_len, rot_vel, osc_len, osc_vel, mult in zip(
            nrgs, rotatory_str_len, rotatory_str_vel, oscillator_str,
            oscillator_vel, multiplicity,
        ):
            self.data.append(
                ValenceExcitation(
                    nrg, rotatory_str_len=rot_len,
                    rotatory_str_vel=rot_vel, oscillator_str=osc_len,
                    oscillator_str_vel=osc_vel, multiplicity=mult,
                )
            )

        for nrg, rot_len, osc_len, osc_vel in zip(
            transient_nrg,
            transient_rot_str,
            transient_oscillator_str,
            transient_oscillator_vel,
        ):
            if not hasattr(self, "transient_data") or not self.transient_data:
                self.transient_data = []
            self.transient_data.append(
                TransientExcitation(
                    nrg,
                    rotatory_str_len=rot_len,
                    oscillator_str=osc_len,
                    oscillator_str_vel=osc_vel,
                )
            )

    def parse_psi4_lines(self, lines, *args, **kwargs):
        symmetry = []
        energy = []
        oscillator_str = []
        oscillator_str_vel = []
        rotation_str = []
        rotation_str_vel = []
        for line in lines:
            if "->" not in line and line.split()[0].isdigit():
                info = line.split()
                symmetry.append(info[1])
                energy.append(float(info[2]))
                oscillator_str.append(float(info[6]))
                rotation_str.append(float(info[7]))
                rotation_str_vel.append(float(info[7]))
                oscillator_str_vel.append(None)
            elif re.search("\| State\s*\d+", line):
                info = re.search("(\d+\.\d+)\s*eV", line)
                energy.append(float(info.group(1)))
            elif re.search("Oscillator strength \(length", line):
                oscillator_str.append(float(line.split()[-1]))
            elif re.search("Oscillator strength \(velocity", line):
                oscillator_str_vel.append(float(line.split()[-1]))
            elif re.search("Rotational strength \(length", line):
                rotation_str.append(float(line.split()[-1]))
            elif re.search("Rotational strength \(velocity", line):
                rotation_str_vel.append(float(line.split()[-1]))
            elif line.split()[0].isdigit():
                info = line[46:].split()
                symmetry.append(line.split("(")[1].split(")")[0])
                energy.append(float(info[0]))
                oscillator_str_vel.append(float(info[2]))
                oscillator_str.append(float(info[3]))
                rotation_str.append(float(info[4]))
                rotation_str_vel.append(float(info[5]))
        
        lists = [
            symmetry, energy, oscillator_str_vel, oscillator_str,
            rotation_str, rotation_str_vel,
        ]
        
        max_list = max(lists, key=len)
        
        for l in lists:
            while len(l) < len(max_list):
                l.append(None)

        for nrg, sym, osc_v, osc, r_l, r_v in zip(
            energy, symmetry, oscillator_str_vel, oscillator_str, 
            rotation_str, rotation_str_vel,
        ):
            self.data.append(
                ValenceExcitation(
                    nrg, symmetry=sym, oscillator_str=osc, rotatory_str_len=r_l,
                    rotatory_str_vel=r_v, oscillator_str_vel=osc_v,
                )
            )

    def parse_qchem_lines(self, lines, *args, **kwargs):
        multiplicity = []
        energy = []
        oscillator_str = []
        symmetry = []
        rotation_str = []
        rotation_str_vel = []
        for line in lines:
            if re.search("Excited state\s+\d+: excitation energy", line):
                energy.append(float(line.split()[-1]))
            
            if re.search("Multiplicity:", line):
                multiplicity.append(line.split()[-1])
            
            if re.search("Strength", line):
                oscillator_str.append(float(line.split()[-1]))
                
            if re.search("Excited state\s+\d+\s*\(", line):
                info = re.search("\((\S+), (\S+)\)", line)
                multiplicity.append(info.group(1).capitalize())
                symmetry.append(info.group(2))
            
            if re.search("Excitation energy:", line):
                if len(energy) > len(oscillator_str):
                    oscillator_str.append(0)
                energy.append(float(line.split()[-2]))

            if re.search("Osc. strength:", line):
                oscillator_str.append(float(line.split()[-1]))
            
            if re.search("State B:", line):
                symmetry.append(line.split("/")[-1])
            
            if re.search("Oscillator strength", line):
                oscillator_str.append(float(line.split()[-1]))

            if re.search("Energy GAP", line):
                energy.append(float(line.split()[-2]))

            if re.search("Rotatory strength, length gauge", line):
                rotation_str.append(float(line.split()[-1]))

            if re.search("Rotatory strength, velocity gauge", line):
                rotation_str_vel.append(float(line.split()[-1]))
        
        lists = [
            symmetry, energy, oscillator_str, multiplicity,
            rotation_str, rotation_str_vel,
        ]
        max_list = max(lists, key=len)
        
        for l in lists:
            while len(l) < len(max_list):
                l.append(None)

        for nrg, mult, osc, symm, rot, rot_vel in zip(
            energy, multiplicity, oscillator_str, symmetry,
            rotation_str, rotation_str_vel,
        ):
            self.data.append(
                ValenceExcitation(
                    nrg, multiplicity=mult, oscillator_str=osc, symmetry=symm,
                    rotatory_str_len=rot, rotatory_str_vel=rot_vel,
                )
            )

    @staticmethod
    def nm_to_ev(x):
        """convert x nm to eV"""
        if isinstance(x, float):
            return PHYSICAL.SPEED_OF_LIGHT * 1e7 * PHYSICAL.PLANCK * UNIT.JOULE_TO_EV / x

        x = np.array(x)
        ndx = np.where(x > 0)
        return PHYSICAL.SPEED_OF_LIGHT * 1e7 * PHYSICAL.PLANCK * UNIT.JOULE_TO_EV / x[ndx], ndx

    @staticmethod
    def ev_to_nm(x):
        """convert x eV to nm"""
        if isinstance(x, float):
            return PHYSICAL.SPEED_OF_LIGHT * 1e7 * PHYSICAL.PLANCK * UNIT.JOULE_TO_EV / x

        x = np.array(x)
        ndx = np.where(x > 0)
        return PHYSICAL.SPEED_OF_LIGHT * 1e7 * PHYSICAL.PLANCK * UNIT.JOULE_TO_EV / x[ndx], ndx

    def plot_uv_vis(
        self,
        figure,
        centers=None,
        widths=None,
        exp_data=None,
        plot_type="uv-vis-veloctiy",
        peak_type="gaussian",
        reverse_x=False,
        y_label=None,
        point_spacing=None,
        normalize=True,
        fwhm=15.0,
        units="nm",
        rotate_x_ticks=False,
        show_functions=None,
        transient=False,
        **kwargs,
    ):
        """
        plot IR data on figure
        figure - matplotlib figure
        centers - array-like of float, plot is split into sections centered
                  on the frequency specified by centers
                  default is to not split into sections
        widths - array-like of float, defines the width of each section
        exp_data - other data to plot
                   should be a list of (x_data, y_data, color)
        reverse_x - if True, 0 cm^-1 will be on the right
        plot_type - see Frequency.get_plot_data
        peak_type - any value allowed by Frequency.get_plot_data
        kwargs - keywords for Frequency.get_spectrum_functions
        """

        data_attr = "data"
        if transient:
            data_attr = "transient_data"
        
        if "intensity_attr" not in kwargs:
            intensity_attr = "oscillator_str"
            if plot_type.lower() == "uv-vis-velocity":
                intensity_attr = "oscillator_str_vel"
            elif plot_type.lower() == "transmittance-velocity":
                intensity_attr = "oscillator_str_vel"
            elif plot_type.lower() == "transmittance":
                intensity_attr = "oscillator_str"
            elif plot_type.lower() == "uv-vis":
                intensity_attr = "oscillator_str"
            elif plot_type.lower() == "ecd":
                intensity_attr = "rotatory_str_len"
            elif plot_type.lower() == "ecd-velocity":
                intensity_attr = "rotatory_str_vel"
            else:
                self.LOG.warning("unrecognized plot type: %s\nDefaulting to uv-vis" % plot_type)
            kwargs["intensity_attr"] = intensity_attr


        if getattr(self.data[0], kwargs["intensity_attr"]) is None:
            raise RuntimeError("no data was parsed for %s" % kwargs["intensity_attr"])

        if not centers and units == "nm":
            data_list = getattr(self, data_attr)
            data_min = None
            data_max = None
            for data in data_list:
                wavelength = self.ev_to_nm(data.excitation_energy)
                if data_min is None or wavelength < data_min:
                    data_min = wavelength
                if data_max is None or wavelength > data_max:
                    data_max = wavelength
            centers = [(data_min + data_max) / 2]
            widths = [1.5 * (data_max - data_min)]

        change_x_unit_func = None
        x_label = "wavelength (nm)"
        change_x_unit_func = self.ev_to_nm
        if units == "eV":
            change_x_unit_func = None
            x_label = r"$h\nu$ (eV)"

        functions, energies, intensities = self.get_spectrum_functions(
            peak_type=peak_type,
            fwhm=fwhm,
            data_attr=data_attr,
            **kwargs,
        )

        other_y_style = None
        ndx_list = None
        if show_functions is not None:
            ndx_list = [info[0] for info in show_functions]
            other_y_style = list(info[1:] for info in show_functions)

        data = self.get_plot_data(
            functions,
            energies,
            fwhm=fwhm,
            transmittance=plot_type.lower().startswith("transmittance"),
            peak_type=peak_type,
            point_spacing=point_spacing,
            change_x_unit_func=change_x_unit_func,
            normalize=normalize,
            show_functions=ndx_list,
        )
        if data is None:
            return

        x_values, y_values, other_y_values = data

        if y_label is None and plot_type.lower().startswith("transmittance"):
            y_label = "Transmittance (%)"
        elif y_label is None and "uv-vis" in plot_type.lower():
            y_label = "Absorbance (arb.)"
        elif y_label is None and "ecd" in plot_type.lower():
            y_label = "ΔAbsorbance (arb.)"

        self.plot_spectrum(
            figure,
            x_values,
            y_values,
            other_y_values=other_y_values,
            other_y_style=other_y_style,
            centers=centers,
            widths=widths,
            exp_data=exp_data,
            reverse_x=reverse_x,
            peak_type=peak_type,
            plot_type=plot_type,
            x_label=x_label,
            y_label=y_label,
            rotate_x_ticks=rotate_x_ticks,
        )



import collections.abc
import os
import re
from collections import OrderedDict
from math import acos, sin, cos, sqrt

import AaronTools.atoms as Atoms
import numpy as np
from AaronTools.const import AARONTOOLS, PHYSICAL


def range_list(number_list, sep=",", sort=True):
    """
    Takes a list of numbers and puts them into a string containing ranges, eg:
    [1, 2, 3, 5, 6, 7, 9, 10] -> "1-3,5-7,9,10"

    :sep: the separator to use between consecutive ranges
    :sort: sort the list before parsing
    """
    if sort:
        number_list = sorted(number_list)
    tmp = [[]]
    for i, n in enumerate(number_list):
        if i == 0:
            tmp[-1] += [n]
        elif n == number_list[i - 1] + 1:
            tmp[-1] += [n]
        else:
            tmp += [[n]]
    rv = ""
    for t in tmp:
        if len(t) > 2:
            rv += "{}-{},".format(t[0], t[-1])
        elif len(t) == 2:
            rv += "{},{},".format(t[0], t[-1])
        else:
            rv += "{},".format(t[0])
    return rv[:-1]


def progress_bar(this, max_num, name=None, width=50):
    if name is None:
        out_str = ""
    else:
        out_str = "{}: ".format(name)
    out_str += "Progress {:3.0f}% ".format(100 * this / max_num)
    out_str += "|{:<{width}s}|".format(
        "#" * int(width * this / max_num), width=width
    )
    print(out_str, end="\r")


def clean_progress_bar(width=50):
    print(" " * 2 * width, end="\r")


def proj(v_vec, u_vec):
    """
    projection of u_vec into v_vec
    v_vec should be a np.array, and u_vec should have the
    same shape as v_vec"""
    numerator = np.dot(u_vec, v_vec)
    denominator = np.linalg.norm(v_vec) ** 2
    return numerator * v_vec / denominator


def quat_matrix(pt1, pt2):
    """build quaternion matrix from pt1 and pt2"""
    pt1 = np.array(pt1, dtype=np.longdouble)
    pt2 = np.array(pt2, dtype=np.longdouble)
    for pt in [pt1, pt2]:
        if len(pt.shape) != 1 or pt.shape[0] != 3:
            raise ValueError("Arguments should be 3-element vectors")

    xm, ym, zm = tuple(pt1 - pt2)
    xp, yp, zp = tuple(pt1 + pt2)

    matrix = np.array(
        [
            [
                xm * xm + ym * ym + zm * zm,
                yp * zm - ym * zp,
                xm * zp - xp * zm,
                xp * ym - xm * yp,
            ],
            [
                yp * zm - ym * zp,
                yp * yp + zp * zp + xm * xm,
                xm * ym - xp * yp,
                xm * zm - xp * zp,
            ],
            [
                xm * zp - xp * zm,
                xm * ym - xp * yp,
                xp * xp + zp * zp + ym * ym,
                ym * zm - yp * zp,
            ],
            [
                xp * ym - xm * yp,
                xm * zm - xp * zp,
                ym * zm - yp * zp,
                xp * xp + yp * yp + zm * zm,
            ],
        ]
    )
    return matrix.astype(np.double)


def uptri2sym(vec, n=None, col_based=False):
    """
    Converts upper triangular matrix to a symmetric matrix

    :vec: the upper triangle array/matrix
    :n: the number of rows/columns
    :col_based: if true, triangular matirx is of the form
                    0 1 3
                    - 2 4
                    - - 5
                if false, triangular matrix is of the form
                    0 1 2
                    - 3 4
                    - - 5
    """
    if hasattr(vec[0], "__iter__") and not isinstance(vec[0], str):
        tmp = []
        for v in vec:
            tmp += v
        vec = tmp
    if n is None:
        n = -1 + np.sqrt(1 + 8 * len(vec))
        n = int(n / 2)
    if n * (n + 1) / 2 != len(vec):
        raise RuntimeError("Bad number of rows requested")

    matrix = np.zeros((n, n))
    if col_based:
        i = 0  # vector index
        j = 0  # for column counting
        for c in range(n):
            j += 1
            for r in range(n):
                matrix[r, c] = vec[i]
                matrix[c, r] = vec[i]
                i += 1
                if r + 1 == j:
                    break
    else:
        for r in range(n):
            for c in range(r, n):
                i = n * r + c - r * (1 + r) / 2
                matrix[r, c] = vec[int(i)]
                matrix[c, r] = vec[int(i)]

    return matrix


def float_vec(word):
    """
    Turns strings into floating point vectors
    :word: a comma-delimited string of numbers

    if no comma or only one element:
        returns just the floating point number
    if elements in word are strings:
        returns word unchanged
    else:
        returns a np.array() of floating point numbers
    """
    val = word
    val = val.split(",")
    try:
        val = [float(v) for v in val]
    except ValueError:
        val = word
    if len(val) == 1:
        return val[0]
    return np.array(val)


def is_alpha(test):
    """determine if string contains only letters"""
    return test.isalpha()


def is_int(test):
    rv = re.search(r"^[+-]?\d+$", test)
    return bool(rv)


def is_num(test):
    """determine if string is valid as a number"""
    rv = re.search(r"^[+-]?\d+\.?\d*", test)
    return bool(rv)


def add_dict(this, other, skip=None):
    if skip is None:
        skip = []
    for key, val in other.items():
        if key in skip:
            continue
        if key in this and isinstance(val, dict):
            add_dict(this[key], val)
        else:
            this[key] = val
    return this


def resolve_concatenation(*args):
    seq = [isinstance(a, collections.abc.MutableSequence) for a in args]
    if any(seq) and not all(seq):
        rv = []
        for i, s in enumerate(seq):
            if s:
                rv.extend(args[i])
            else:
                rv.append(args[i])
        return rv

    err_msg = "Cannot concatenate" + " {}" * len(args)
    raise TypeError(err_msg.format(*[type(a) for a in args]))


def combine_dicts(*args, case_sensitive=False, dict2_conditional=False):
    """combine dictionaries d1 and d2 to return a dictionary
    with keys d1.keys() + d2.keys()
    if a key is in d1 and d2, the items will be combined:
        if they are both dictionaries, combine_dicts is called recursively
        otherwise, d2[key] is appended to d1[key]
        if case_sensitive=False, the key in the output will be the lowercase
        of the d1 key and d2 key (only for combined items)
    dict2_conditional: bool - if True, don't add d2 keys unless they are also in d1
    """
    from copy import deepcopy

    d1 = args[0]
    d2 = args[1:]
    if len(d2) > 1:
        d2 = combine_dicts(
            d2[0],
            *d2[1:],
            case_sensitive=case_sensitive,
            dict2_conditional=dict2_conditional
        )
    else:
        d2 = d2[0]

    out = OrderedDict()
    case_keys_1 = list(d1.keys())
    case_keys_2 = list(d2.keys())
    if case_sensitive:
        keys_1 = case_keys_1
        keys_2 = case_keys_2
    else:
        keys_1 = [
            key.lower() if isinstance(key, str) else key for key in case_keys_1
        ]
        keys_2 = [
            key.lower() if isinstance(key, str) else key for key in case_keys_2
        ]

    # go through keys from d1
    for case_key, key in zip(case_keys_1, keys_1):
        # if the key is only in d1, add the item to out
        if key in keys_1 and key not in keys_2:
            out[case_key] = deepcopy(d1[case_key])
        # if the key is in both, combine the items
        elif key in keys_1 and key in keys_2:
            key_2 = case_keys_2[keys_2.index(key)]
            if isinstance(d1[case_key], dict) and isinstance(d2[key_2], dict):
                out[key] = combine_dicts(
                    d1[case_key],
                    d2[key_2],
                    case_sensitive=case_sensitive,
                )
            else:
                try:
                    out[key] = deepcopy(d1[case_key]) + deepcopy(d2[key_2])
                except TypeError:
                    out[key] = resolve_concatenation(d1[case_key], d2[key_2])

    # go through keys from d2
    if not dict2_conditional:
        for case_key, key in zip(case_keys_2, keys_2):
            # if it's only in d2, add item to out
            if key in keys_2 and key not in keys_1:
                out[case_key] = d2[case_key]

    return out


def integrate(fun, start, stop, num=101):
    """numerical integration using Simpson's method
    fun - function to integrate
    start - starting point for integration
    stop - stopping point for integration
    num - number of points used for integration"""

    delta_x = float(stop - start) / (num - 1)
    x_set = np.linspace(start, stop, num=num)
    running_sum = -(fun(start) + fun(stop))
    i = 1
    max_4th_deriv = 0
    while i < num:
        if i % 2 == 0:
            running_sum += 2 * fun(x_set[i])
        else:
            running_sum += 4 * fun(x_set[i])

        if i < num - 4 and i >= 3:
            sg_4th_deriv = (
                6 * fun(x_set[i - 3])
                + 1 * fun(x_set[i - 2])
                - 7 * fun(x_set[i - 1])
                - 3 * fun(x_set[i])
                - 7 * fun(x_set[i + 1])
                + fun(x_set[i + 2])
                + 6 * fun(x_set[i + 3])
            )
            sg_4th_deriv /= 11 * delta_x ** 4

            if abs(sg_4th_deriv) > max_4th_deriv:
                max_4th_deriv = abs(sg_4th_deriv)

        i += 1

    running_sum *= delta_x / 3.0

    # close enough error estimate
    e = (abs(stop - start) ** 5) * max_4th_deriv / (180 * num ** 4)

    return (running_sum, e)


def same_cycle(graph, a, b):
    """
    Determines if Atom :a: and Atom :b: are in the same cycle in a undirected :graph:

    Returns: True if cycle found containing a and b, False otherwise

    :graph: connectivity matrix or Geometry
    :a:, :b: indices in connectivity matrix/Geometry or Atoms in Geometry
    """
    from AaronTools.geometry import Geometry

    if isinstance(a, Atoms.Atom):
        a = graph.atoms.index(a)
    if isinstance(b, Atoms.Atom):
        b = graph.atoms.index(b)
    if isinstance(graph, Geometry):
        graph = [
            [graph.atoms.index(j) for j in i.connected] for i in graph.atoms
        ]
    graph = [[i for i in j] for j in graph]

    graph, removed = trim_leaves(graph)

    if a in removed or b in removed:
        return False
    path = shortest_path(graph, a, b)
    for p, q in zip(path[:-1], path[1:]):
        graph[p].remove(q)
        graph[q].remove(p)

    path = shortest_path(graph, a, b)
    if path is None:
        return False
    return True


def shortest_path(graph, start, end):
    """
    Find shortest path from :start: to :end: in :graph: using Dijkstra's algorithm
    Returns: list(node_index) if path found, None if path not found

    :graph: the connection matrix or Geometry
    :start: the first atom or node index
    :end: the last atom or node index
    """
    from AaronTools.geometry import Geometry

    if isinstance(start, Atoms.Atom):
        start = graph.atoms.index(start)
    if isinstance(end, Atoms.Atom):
        end = graph.atoms.index(end)
    if isinstance(graph, Geometry):
        graph = [
            [graph.atoms.index(j) for j in i.connected if j in graph.atoms]
            for i in graph.atoms
        ]
    graph = [[i for i in j] for j in graph]

    # initialize distance array, parent array, and set of unvisited nodes
    dist = [np.inf for x in graph]
    parent = [-1 for x in graph]
    unvisited = set([i for i in range(len(graph))])

    dist[start] = 0
    current = start
    while True:
        # for all unvisited neighbors of current node, update distances
        # if we update the distance to a neighboring node,
        # then also update its parent to be the current node
        for v in graph[current]:
            if v not in unvisited:
                continue
            if dist[v] == np.inf:
                new_dist = dist[current] + 1
            else:
                new_dist = dist[current] + dist[v]
            if dist[v] > new_dist:
                dist[v] = new_dist
                parent[v] = current
        # mark current node as visited
        # select closest unvisited node to be next node
        # break loop if we found end node or if no unvisited connected nodes
        unvisited.remove(current)
        if end not in unvisited:
            break
        current = None
        for u in unvisited:
            if current is None or dist[u] < dist[current]:
                current = u
        if dist[current] == np.inf:
            break
    # return shortest path from start to end
    path = [end]
    while True:
        if parent[path[-1]] == -1:
            break
        path += [parent[path[-1]]]
    path.reverse()
    if path[0] != start or path[-1] != end:
        return None
    return path


def trim_leaves(graph, _removed=None):
    from AaronTools.geometry import Geometry

    # print(_removed)
    if _removed is None:
        _removed = []

    if isinstance(graph, Geometry):
        graph = [
            [graph.atoms.index(j) for j in i.connected] for i in graph.atoms
        ]
    graph = [[i for i in j] for j in graph]
    some_removed = False

    for i, con in enumerate(graph):
        if len(con) == 1:
            graph[con[0]].remove(i)
            graph[i].remove(con[0])
            some_removed = True
            _removed += [i]

    if some_removed:
        graph, _removed = trim_leaves(graph, _removed)

    return graph, set(_removed)


def to_closing(string, brace):
    """returns the portion of string from the beginning to the closing
    paratheses or bracket denoted by brace
    brace can be '(', '{', or '['
    if the closing paratheses is not found, returns None instead"""
    if brace == "(":
        pair = ("(", ")")
    elif brace == "{":
        pair = ("{", "}")
    elif brace == "[":
        pair = ("[", "]")
    else:
        raise RuntimeError("brace must be '(', '{', or '['")

    out = ""
    count = 0
    for x in string:
        if x == pair[0]:
            count += 1
        elif x == pair[1]:
            count -= 1

        out += x
        if count == 0:
            break

    if count != 0:
        return None
    else:
        return out


def rotation_matrix(theta, axis, renormalize=True):
    """rotation matrix for rotating theta radians about axis"""
    # I've only tested this for rotations in R3
    dim = len(axis)
    if renormalize:
        if np.linalg.norm(axis) == 0:
            axis = np.zeros(dim)
            axis[0] = 1.0
        axis = axis / np.linalg.norm(axis)
    outer_prod = np.outer(axis, axis)
    cos_comp = cos(theta)
    outer_prod *= 1 - cos_comp
    iden = np.identity(dim)
    cos_comp = iden * cos_comp
    sin_comp = sin(theta) * (np.ones((dim, dim)) - iden)
    cross_mat = np.zeros((dim, dim))
    for i in range(0, dim):
        for j in range(0, i):
            p = 1
            if (i + j) % 2 != 0:
                p = -1

            cross_mat[i][j] = -1 * (p * axis[dim - (i + j)])
            cross_mat[j][i] = p * axis[dim - (i + j)]

    return outer_prod + cos_comp + sin_comp * cross_mat


def mirror_matrix(norm, renormalize=True):
    """mirror matrix for the specified norm"""
    if renormalize:
        norm /= np.linalg.norm(norm)
    A = np.identity(3)
    B = -2 * np.outer(norm, norm)
    return A + B


def fibonacci_sphere(radius=1, center=np.zeros(3), num=500):
    """
    returns a grid of points that are equally spaced on a sphere
    with the specified radius and center
    number of points can be adjusted with num
    """
    # generate a grid of points on the unit sphere
    grid = np.zeros((num, 3))
    
    i = np.arange(0, num) + 0.5
    phi = np.arccos(1 - 2 * i / num)
    ratio = (1 + sqrt(5.)) / 2
    theta = 2 * np.pi * i / ratio
    grid[:, 0] = np.cos(theta) * np.sin(phi)
    grid[:, 1] = np.sin(theta) * np.sin(phi)
    grid[:, 2] = np.cos(phi)

    # scale the points to the specified radius and move the center
    grid *= radius
    grid += center

    return grid


def lebedev_sphere(radius=1, center=np.zeros(3), num=302):
    """
    returns one of the Lebedev grid points (xi, yi, zi)
    and weights (wi) with the specified radius and center.
    Weights do not include r**2, so integral of F(x,y,z)
    over sphere is 4*pi*r**2\sum_i{F(xi,yi,zi)wi}.  The number
    of points (num) must be one of 110, 194, 302, 590, 974, 1454, 2030, 2702, 5810
    """
    # read grid data  on unit sphere
    grid_file = os.path.join(
        AARONTOOLS, "utils", "quad_grids", "Leb" + str(num) + ".grid"
    )
    if not os.path.exists(grid_file):
        # maybe some other error type?
        raise NotImplementedError(
            "cannot use Lebedev grid with %i points\n" % num
            + "use one of 110, 194, 302, 590, 974, 1454, 2030, 2702, 5810"
        )
    grid_data = np.loadtxt(grid_file)
    grid = grid_data[:, [0, 1, 2]]
    weights = grid_data[:, 3]

    # scale the points to the specified radius and move the center
    grid *= radius
    grid += center

    return grid, weights


def gauss_legendre_grid(start=-1, stop=1, num=32):
    """
    returns a Gauss-Legendre grid points (xi) and weights
    (wi)for the range start to stop. Integral over F(x) is
    \sum_i{F(xi)wi}. The number of points (num) must be one
    of 20, 32, 64, 75, 99, 127
    """
    # read grid points on the range [-1,1] and weights
    grid_file = os.path.join(
        AARONTOOLS, "utils", "quad_grids", "Leg" + str(num) + ".grid"
    )
    if not os.path.exists(grid_file):
        # maybe some other error type?
        raise NotImplementedError(
            "cannot use Gauss-Legendre grid with %i points\n" % num
            + "use one of 20, 32, 64, 75, 99, 127"
        )
    grid_data = np.loadtxt(grid_file)

    # shift grid range to [start, stop]
    grid = grid_data[:, 0] * (stop - start) / 2 + start + (stop - start) / 2

    # adjust weights for new range
    weights = grid_data[:, 1] * (stop - start) / 2

    return grid, weights


def perp_vector(vec):
    """
    returns a vector orthonormal to vec (np.ndarray)
    if vec is 2D, returns a vector orthonormal to the
    plane of best for the rows of vec
    """
    vec = np.squeeze(vec)
    if vec.ndim == 1:
        out_vec = np.roll(vec, 1)
        if all(x == vec[0] for x in vec):
            for k in range(0, len(vec)):
                if out_vec[k] != 0:
                    out_vec[k] *= -1
                    break
        if np.linalg.norm(vec) == 0:
            # a zero-vector was given
            return np.ones(len(vec)) / len(vec)

        out = np.cross(out_vec, vec)
        out /= np.linalg.norm(out)
        return out

    elif vec.ndim == 2:
        xyz = vec - np.mean(vec, axis=0)
        cov_prod = np.dot(xyz.T, xyz)
        u, s, vh = np.linalg.svd(cov_prod, compute_uv=True)
        return u[:, -1]

    raise NotImplementedError(
        "cannot determine vector perpendicular to %i-dimensional array"
        % vec.ndim
    )


def get_filename(path, include_parent_dir=True):
    """returns the name of the file without parent directories or extension"""
    if not include_parent_dir:
        fname = os.path.basename(path)
    else:
        fname = path
    fname, _ = os.path.splitext(fname)
    return fname


def boltzmann_coefficients(energies, temperature, absolute=True):
    """
    returns boltzmann weights for the energies and T
    energies - numpy array of energies in kcal/mol
    temperature - T in K
    absolute - True if the energies given are absolute, false if they are relative energies
    """
    if absolute:
        min_nrg = min(energies)
        energies -= min_nrg
    weights = np.exp(-energies / (PHYSICAL.R * temperature))
    return weights


def boltzmann_average(energies, values, temperature, absolute=True):
    """
    returns the AVT result for the values corresponding to the energies
    energies - np.array, energy for each state in kcal/mol
    values - np.array, values for which the weighting is applied; the ith value corresponds to the ith energy
    temperature - float, temperature in K
    absolute - True if the energies given are absolute, false if they are relative energies
    """
    weights = boltzmann_coefficients(energies, temperature, absolute=absolute)
    avg = np.dot(weights, values) / sum(weights)
    return avg


def glob_files(infiles, parser=None):
    """
    globs input files
    used for command line scripts because Windows doesn't support globbing...
    """
    from glob import glob
    import sys

    if isinstance(infiles, str):
        infiles = [infiles]

    outfiles = []
    for f in infiles:
        if isinstance(f, str):
            outfiles.extend(glob(f))
        elif len(sys.argv) > 1:
            outfiles.append(f)

    if not outfiles and all(isinstance(f, str) for f in infiles):
        raise RuntimeError(
            "no files could be found for %s" % ", ".join(infiles)
        )
    elif not outfiles and parser is not None:
        parser.print_help()
        sys.exit(0)

    return outfiles


def angle_between_vectors(v1, v2, renormalize=True):
    """returns the angle between v1 and v2 (numpy arrays)"""
    if renormalize:
        v1 = v1 / np.linalg.norm(v1)
        v2 = v2 / np.linalg.norm(v2)

    v12 = v2 - v1
    c2 = np.dot(v12, v12)
    t = (c2 - 2) / -2
    if t > 1:
        t = 1
    elif t < -1:
        t = -1

    # math.acos is faster than numpy.arccos for non-arrays
    return acos(t)


def is_alpha(test):
    rv = re.search("^[a-zA-Z]+$", test)
    return bool(rv)


def is_int(test):
    rv = re.search("^[+-]?\d+$", test)
    return bool(rv)


def is_num(test):
    rv = re.search("^[+-]?\d+\.?\d*", test)
    return bool(rv)

float_num = re.compile("[-+]?\d+\.?\d*")



#!/usr/bin/env python3

import argparse
import os
import re
import sys
from numpy import prod

from AaronTools.fileIO import FileReader, read_types
from AaronTools.finders import BondedTo
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.utils.utils import get_filename, glob_files

makeconf_parser = argparse.ArgumentParser(
    description="generate rotamers for substituents using a hierarchical method",
    formatter_class=argparse.RawTextHelpFormatter
)

makeconf_parser.add_argument(
    "infile", metavar="input file",
    type=str,
    nargs="*",
    default=[sys.stdin],
    help="a coordinate file"
)

makeconf_parser.add_argument(
    "-ls", "--list",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="list_avail",
    help="list available substituents"
)

makeconf_parser.add_argument(
    "-i", "--info",
    action="store_const",
    const=True,
    default=False,
    required=False,
    dest="list_info",
    help="list information on substituents to be generated"
)

makeconf_parser.add_argument(
    "-if", "--input-format",
    type=str,
    nargs=1,
    default=None,
    choices=read_types,
    dest="input_format",
    help="file format of input - xyz is assumed if input is stdin"
)

makeconf_parser.add_argument(
    "-s", "--substituent", metavar="(n=substituent|substituent name)",
    type=str,
    action="append",
    default=[],
    required=False,
    dest="substituents",
    help="substituents to rotate \n" +
    "n is the 1-indexed position of the starting position of the\n" +
    "substituent you are rotating\n" +
    "if only a substituent name is specified, all substituents with\n" +
    "that name will be rotated\n" +
    "Default: rotate any detected substituents"
)

makeconf_parser.add_argument(
    "-rc", "--remove-clash",
    action="store_true",
    default=False,
    required=False,
    dest="remove_clash",
    help="if atoms are too close together, wiggle the\n" +
    "substituents to remove the clash"
)

makeconf_parser.add_argument(
    "-sc", "--skip-clash",
    action="store_true",
    required=False,
    default=False,
    dest="skip_clash",
    help="do not print structures with atoms that are\n" +
    "too close together or for which substituent\n" +
    "clashing could not be resolved with '--remove-clash'"
)

makeconf_parser.add_argument(
    "-o", "--output-destination",
    type=str,
    default=None,
    required=False,
    metavar="output destination",
    dest="outfile",
    help="output destination\n" +
    "$i in the filename will be replaced with conformer number\n" +
    "if a directory is given, default is \"conformer-$i.xyz\" in that directory\n" +
    "$INFILE will be replaced with the name of the input file\n" +
    "Default: stdout"
)

args = makeconf_parser.parse_args()

if args.list_avail:
    s = ""
    for i, name in enumerate(sorted(Substituent.list())):
        sub = Substituent(name)
        if sub.conf_num > 1:
            s += "%-20s" % name
            # if (i + 1) % 3 == 0:
            if (i + 1) % 1 == 0:
                s += "\n"

    print(s.strip())
    sys.exit(0)

detect_subs = False

s = ""

skipped = 0

for infile in glob_files(args.infile, parser=makeconf_parser):
    if isinstance(infile, str):
        if args.input_format is not None:
            f = FileReader((infile, args.input_format[0], infile))
        else:
            f = FileReader(infile)
    else:
        if args.input_format is not None:
            f = FileReader(("from stdin", args.input_format[0], infile))
        else:
            f = FileReader(("from stdin", "xyz", infile))

    geom = Geometry(f)

    target_list = []
    explicit_subnames = []
    subnames = []
    if not args.substituents:
        detect_subs = True
    else:
        for sub in args.substituents:
            if re.search(r"^\d+.*=\w+", sub):
                ndx_targets = sub.split("=")[0]
                target_list.append(geom.find(ndx_targets))
                explicit_subnames.append("=".join(sub.split("=")[1:]))
            else:
                detect_subs = True
                subnames.append(sub)

    substituents = []
    if detect_subs:
        geom.detect_substituents()
        if not subnames:
            substituents.extend(
                [sub for sub in geom.substituents if sub.name in Substituent.list()]
            )
        else:
            substituents.extend([sub for sub in geom.substituents if sub.name in subnames])

    for a, subname in zip(target_list, explicit_subnames):
        for atom in a:
            for bonded_atom in atom.connected:
                frag = geom.get_fragment(atom, bonded_atom)
                try:
                    sub = Substituent(frag, end=bonded_atom)
                    if sub.name == subname:
                        substituents.append(sub)

                except LookupError:
                    pass

    if args.list_info:
        total_conf = 1
        if len(args.infile) > 1:
            s += "%s\n" % infile
        else:
            s += ""
        s += "Substituent \tRotamers\n"
        for sub in substituents:
            if sub.conf_num > 1:
                s += "%2s=%-10s\t%s\n" % (sub.end.name, sub.name, sub.conf_num)
                total_conf *= sub.conf_num
        s += "Total Number of Conformers = %i\n" % total_conf
        if infile is not args.infile[-1]:
            s += "\n"
        continue

    # imagine conformers as a number
    # each place in that number is in base conf_num
    # we determine what to rotate by adding one (starting from 0) and subtracting the
    # previous number
    conformers = []
    rotations = []
    for sub in substituents:
        conformers.append(sub.conf_num)
        rotations.append(sub.conf_angle)

    mod_array = []
    for i in range(0, len(rotations)):
        mod_array.append(1)
        for j in range(0, i):
            mod_array[i] *= conformers[j]

    prev_conf = 0
    for conf in range(0, int(prod(conformers))):
        for i, sub in enumerate(substituents):
            rot = int(conf / mod_array[i]) % conformers[i]
            rot -= int(prev_conf / mod_array[i]) % conformers[i]
            angle = rotations[i] * rot
            if angle != 0:
                sub_atom = sub.find_exact(BondedTo(sub.end))[0]
                axis = sub_atom.bond(sub.end)
                center = sub.end.coords
                geom.rotate(axis, angle=angle, targets=sub.atoms, center=center)

        prev_conf = conf

        bad_subs = []
        print_geom = geom
        if args.remove_clash:
            print_geom = Geometry([a.copy() for a in geom])
            # print_geom.update_geometry(geom.coordinates.copy())
            sub_list = [Substituent(print_geom.find([at.name for at in sub]),
                                    detect=False,
                                    end=print_geom.find_exact(sub.end.name)[0])
                        for sub in substituents]
            bad_subs = print_geom.remove_clash(sub_list)
            # somehow the atoms get out of order
            print_geom.atoms = sorted(print_geom.atoms, key=lambda a: float(a.name))

        if args.skip_clash:
            if bad_subs:
                skipped += 1
                continue

            clashing = False
            for j, a1 in enumerate(print_geom.atoms):
                for a2 in print_geom.atoms[:j]:
                    if a2 not in a1.connected and a1.is_connected(a2):
                        clashing = True
                        break

                if clashing:
                    break

            if clashing:
                skipped += 1
                continue

        if args.outfile is None:
            s += print_geom.write(outfile=False)
            s += "\n"
        else:
            if os.path.isdir(os.path.expanduser(args.outfile)):
                outfile = os.path.join(
                    os.path.expanduser(args.outfile), "conformer-%i.xyz" % (conf + 1)
                )

            else:
                outfile = args.outfile.replace("$i", str(conf + 1))

            if "$INFILE" in outfile:
                outfile = outfile.replace("$INFILE", get_filename(infile))

            print_geom.write(outfile=outfile, append="$i" not in args.outfile)


if args.outfile is None or args.list_info:
    print(s[:-1])

if skipped > 0:
    print("%i conformers skipped because of clashing substituent(s)" % skipped)



#!/usr/bin/env python3
import os
import unittest

from AaronTools.geometry import Geometry
from AaronTools.test import TestWithTimer, prefix
from AaronTools.symmetry import PointGroup

class TestSymmetry(TestWithTimer):
    benzene = os.path.join(prefix, "test_files", "benzene.xyz")
    c60 = os.path.join(prefix, "test_files", "c60.xyz")
    adamantane = os.path.join(prefix, "test_files", "adamantane.xyz")
    methane = os.path.join(prefix, "test_files", "methane.xyz")
    c3h4 = os.path.join(prefix, "test_files", "c3h4.xyz")
    h2o = os.path.join(prefix, "test_files", "h2o.xyz")
    meoh = os.path.join(prefix, "test_files", "MeOH.xyz")
    Me_NO2_4 = os.path.join(prefix, "test_files", "C(NO2)4.xyz")
    d4h = os.path.join(prefix, "test_files", "d4h.xyz")
    ci = os.path.join(prefix, "test_files", "Ci.xyz")
    oh = os.path.join(prefix, "test_files", "cubane.xyz")
    d5h = os.path.join(prefix, "test_files", "c5h5.xyz")
    t = os.path.join(prefix, "test_files", "neopentane.xyz")
    cubeamine = os.path.join(prefix, "test_files", "cubeamine.xyz")

    chiral_ring = os.path.join(prefix, "test_files", "chiral_ring.xyz")
    chiral_mol_1 = os.path.join(prefix, "test_files", "chiral_centers_1.xyz")
    chiral_mol_2 = os.path.join(prefix, "test_files", "chiral_centers_2.xyz")
    chiral_mol_3 = os.path.join(prefix, "test_files", "chiral_centers_3.xyz")
    chiral_mol_4 = os.path.join(prefix, "test_files", "chiral_centers_4.xyz")

    def test_D6h(self):
        # test to see if interpolated geometry is correct
        mol = Geometry(self.benzene, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "D6h")

    def test_Cs(self):
        # test to see if interpolated geometry is correct
        mol = Geometry(self.meoh, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "Cs")

    def test_S4(self):
        # test to see if interpolated geometry is correct
        mol = Geometry(self.Me_NO2_4, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "S4")

    def test_C1(self):
        mol1 = Geometry(self.chiral_ring, refresh_ranks=False)
        pg = PointGroup(mol1)

        self.assertEqual(pg.name, "C1")

        mol2 = Geometry(self.chiral_mol_1, refresh_ranks=False)
        pg = PointGroup(mol2)

        self.assertEqual(pg.name, "C1")

        mol3 = Geometry(self.chiral_mol_2, refresh_ranks=False)
        pg = PointGroup(mol3)

        self.assertEqual(pg.name, "C1")

        mol4 = Geometry(self.chiral_mol_3, refresh_ranks=False)
        pg = PointGroup(mol4)

        self.assertEqual(pg.name, "C1")

        mol5 = Geometry(self.chiral_mol_4, refresh_ranks=False)
        pg = PointGroup(mol5)

        self.assertEqual(pg.name, "C1")

    def test_Ih(self):
        mol = Geometry(self.c60, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "Ih")

    def test_Td(self):
        mol = Geometry(self.adamantane, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "Td")

        mol = Geometry(self.methane, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "Td")

    def test_D2d(self):
        mol = Geometry(self.c3h4, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "D2d")

    def test_C2v(self):
        mol = Geometry(self.h2o, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "C2v")

    def test_D4h(self):
        mol = Geometry(self.d4h, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "D4h")

    def test_Ci(self):
        mol = Geometry(self.ci, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "Ci")

    def test_Oh(self):
        mol = Geometry(self.oh, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "Oh")

    def test_D5h(self):
        mol = Geometry(self.d5h, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "D5h")

    def test_T(self):
        mol = Geometry(self.t, refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "T")

    def test_O(self):
        mol = Geometry(self.cubeamine, refresh_ranks=False, refresh_connected=False)
        mol = Geometry(mol.find(["C", "N"]), refresh_ranks=False)
        pg = PointGroup(mol)

        self.assertEqual(pg.name, "O")

def suite():
    suite = unittest.TestSuite()
    suite.addTest(TestSymmetry("test_Ih"))
    suite.addTest(TestSymmetry("test_Oh"))
    suite.addTest(TestSymmetry("test_O"))
    suite.addTest(TestSymmetry("test_Td"))
    suite.addTest(TestSymmetry("test_D6h"))
    suite.addTest(TestSymmetry("test_D5h"))
    suite.addTest(TestSymmetry("test_D4h"))
    suite.addTest(TestSymmetry("test_D2d"))
    suite.addTest(TestSymmetry("test_S4"))
    suite.addTest(TestSymmetry("test_C2v"))
    suite.addTest(TestSymmetry("test_Cs"))
    suite.addTest(TestSymmetry("test_Ci"))
    suite.addTest(TestSymmetry("test_C1"))
    return suite


if __name__ == "__main__":
    # import cProfile

    # profile = cProfile.Profile()
    # profile.enable()
    runner = unittest.TextTestRunner()
    runner.run(suite())
    # profile.disable()
    # profile.print_stats()



#!/usr/bin/env python3

import sys
import argparse

from AaronTools.geometry import Geometry

fetch_parser = argparse.ArgumentParser(
    description="print structure in xyz format",
    formatter_class=argparse.RawTextHelpFormatter
)

fetch_parser.add_argument(
    "-o", "--output",
    type=str,
    default=False,
    required=False,
    dest="outfile",
    help="output destination \nDefault: stdout"
)

input_format = fetch_parser.add_argument_group("Identifier format")
style = input_format.add_mutually_exclusive_group(required=True)
style.add_argument(
    "-s", "--smiles",
    type=str,
    default=None,
    required=False,
    dest="smiles",
    help="SMILES notation for molecule"
)

style.add_argument(
    "-i", "--iupac",
    type=str,
    default=None,
    required=False,
    dest="iupac",
    help="IUPAC name of molecule"
)

args = fetch_parser.parse_args()

if args.smiles is not None:
    geom = Geometry.from_string(args.smiles, form="smiles")
elif args.iupac is not None:
    geom = Geometry.from_string(args.iupac, form="iupac")

if not args.outfile:
    print(geom.write(outfile=False))
else:
    geom.write(
        append=True, 
        outfile=args.outfile
    )



#!/usr/bin/env python3
import os
import unittest

import numpy as np
from AaronTools.fileIO import FileReader, FileWriter
from AaronTools.geometry import Geometry
from AaronTools.test import TestWithTimer, prefix
from AaronTools.theory import *


class TestFileReader(TestWithTimer):
    small_mol = os.path.join(prefix, "test_files", "benzene_1-NO2_4-Cl.xyz")
    com_file1 = os.path.join(prefix, "test_files", "test-route.com")
    com_file2 = os.path.join(prefix, "test_files", "test-route-2.com")
    psi4_output_file = os.path.join(prefix, "test_files", "psi4-test.out")
    orca_orbit_file = os.path.join(prefix, "test_files", "pople.out")
    gaussian_mp2_file = os.path.join(prefix, "test_files", "mp2.log")
    gaussian_mp4_file = os.path.join(prefix, "test_files", "mp4.log")
    gaussian_ccsd_file = os.path.join(prefix, "test_files", "ccsd.log")
    gaussian_ccsd_t_file = os.path.join(prefix, "test_files", "ccsd_t.log")
    gaussian_dhdft_file = os.path.join(prefix, "test_files", "b2plypd3.log")
    ts = os.path.join(prefix, "test_files", "claisen_ts.xyz")

    def xyz_matrix(self, fname):
        rv = []
        with open(fname) as f:
            rv += [f.readline().strip()]
            rv += [f.readline().strip()]
            for line in f:
                line = line.strip()
                if len(line) == 0:
                    continue
                rv += [line.split()]
        return rv

    def validate_atoms(self, ref, test, skip=[]):
        for r, t in zip(ref.atoms, test.atoms):
            if r.element != t.element:
                return False
            if np.linalg.norm(r.coords - t.coords) > 10 ** -5:
                return False
            if "name" not in skip and r.name != t.name:
                return False
            if "flag" not in skip and r.flag != t.flag:
                return False
            if "tags" not in skip:
                if len(r.tags - t.tags) > 0:
                    return False
            return True

    def test_read_xyz_structure(self):
        ref = self.xyz_matrix(TestFileReader.small_mol)
        mol = FileReader(TestFileReader.small_mol)
        for i, line in enumerate(ref):
            if i == 0:
                # check number of atoms
                self.assertEqual(len(mol.atoms), int(line))
                continue
            if i == 1:
                # check comment
                self.assertEqual(mol.comment, line)
                continue
            # check element and coordinates
            n = i - 2  # atoms start at line 2
            a = mol.atoms[n]
            self.assertEqual(a.element, line[0])
            dist = a.coords - np.array(line[1:], dtype=float)
            dist = np.linalg.norm(dist)
            self.assertTrue(dist < 10 ** (-8))
            # check name and tags
            self.assertEqual(str(n + 1), a.name)
            self.assertTrue(len(a.tags) == 0)

    def test_read_orca_out_structure(self):
        ref = FileReader(os.path.join(prefix, "ref_files", "orca_geom.xyz"))
        test = FileReader(os.path.join(prefix, "test_files", "orca_geom.out"))
        self.assertTrue(self.validate_atoms(ref, test))

    def test_read_orca_orbits(self):
        # this file can be tricky b/c MO coefficients are right next
        # to each other - no spaces between them
        test = FileReader(self.orca_orbit_file, just_geom=False)
        self.assertEqual(test.other["orbitals"].n_mos, 168)

    def test_read_psi4_dat_structure(self):
        ref = FileReader(os.path.join(prefix, "ref_files", "psi4_geom.xyz"))
        test = FileReader((self.psi4_output_file, "dat", None))
        self.assertTrue(self.validate_atoms(ref, test))

    def test_read_log_structure(self):
        ref = FileReader(
            os.path.join(prefix, "ref_files", "file_io_normal.xyz")
        )
        test = FileReader(os.path.join(prefix, "test_files", "normal.log"))
        self.assertTrue(self.validate_atoms(ref, test))

        ref = FileReader(
            os.path.join(prefix, "ref_files", "file_io_error.xyz")
        )
        test = FileReader(os.path.join(prefix, "test_files", "error.log"))
        self.assertTrue(self.validate_atoms(ref, test))

        ref = FileReader(os.path.join(prefix, "ref_files", "file_io_died.xyz"))
        test = FileReader(os.path.join(prefix, "test_files", "died.log"))
        self.assertTrue(self.validate_atoms(ref, test))

    def test_read_log_energies(self):
        """reading the correct energy from Gaussian output"""
        # mp2
        fr = FileReader(self.gaussian_mp2_file, just_geom=False)
        self.assertEqual(fr.other["energy"], -76.228479251471)
        
        # mp4
        fr = FileReader(self.gaussian_mp4_file, just_geom=False)
        self.assertEqual(fr.other["energy"], -76.240714654)
        
        # ccsd
        fr = FileReader(self.gaussian_ccsd_file, just_geom=False)
        self.assertEqual(fr.other["energy"], -76.238041859)
        
        # ccsd(t)
        fr = FileReader(self.gaussian_ccsd_t_file, just_geom=False)
        self.assertEqual(fr.other["energy"], -76.241082564)
        
        # double hybrid dft
        fr = FileReader(self.gaussian_dhdft_file, just_geom=False)
        self.assertEqual(fr.other["energy"], -76.353361915801)

    def test_read_com_info(self):
        """testing if we can read route info"""
        ref1 = {
            "method": "B3LYP/aug-cc-pVDZ",
            "temperature": "298.15",
            "solvent": "1,1,1-TriChloroEthane",
            "solvent_model": "PCM",
            "emp_dispersion": "GD3",
            "grid": "SuperFineGrid",
            "comment": "testing 1 2 3\ntesting 1 2 3",
            "charge": 0,
            "multiplicity": 1,
        }

        test = FileReader(self.com_file1, just_geom=False)
        self.assertEqual(test.other, ref1)

        ref2 = {
            "method": "B3LYP/aug-cc-pVDZ",
            "temperature": "298.15",
            "solvent": "1,1,1-TriChloroEthane",
            "solvent_model": "PCM",
            "emp_dispersion": "GD3",
            "grid": "ultrafinegrid",
            "comment": "testing 1 2 3\ntesting 1 2 3",
            "charge": 0,
            "multiplicity": 1,
        }

        test = FileReader(self.com_file2, just_geom=False)
        self.assertEqual(test.other, ref2)

    def test_write_com(self):
        """write gaussian input file"""
        # this compares the exact string, not things like RMSD or dictionaries
        # if it fails, someone may have added a column of whitespace or something
        geom = Geometry(self.small_mol)

        ref = """#n PBE1PBE/gen freq=(temperature=298.15,HPModes,NoRaman) opt=(VeryTight) EmpiricalDispersion=(GD3BJ) scrf=(SMD,solvent=dichloromethane)

comment line 1
comment line 2

0 1
C     -1.97696    -2.32718     0.00126
C     -2.36814    -1.29554     0.85518
C     -1.67136    -0.08735     0.85440
C     -0.58210     0.08919     0.00026
C     -0.19077    -0.94241    -0.85309
C     -0.88848    -2.15056    -0.85289
H     -3.22679    -1.43483     1.52790
H     -1.98002     0.72606     1.52699
H      0.66766    -0.80358    -1.52636
H     -0.57992    -2.96360    -1.52585
Cl     0.29699     1.61392    -0.00037
N     -2.73689    -3.64357     0.00188
O     -2.07823    -4.68230     0.00289
O     -3.96579    -3.59263     0.00134

H 0
def2SVP
****
C 0
def2TZVP
****



"""

        theory = Theory(
            charge=0,
            multiplicity=1,
            method="PBE0",
            basis=BasisSet(
                [Basis("def2-SVP", ["H"]), Basis("def2-TZVP", ["C"])]
            ),
            empirical_dispersion=EmpiricalDispersion("D3BJ"),
            solvent=ImplicitSolvent("SMD", "dichloromethane"),
            job_type=[FrequencyJob(), OptimizationJob()],
        )

        kw_dict = {
            GAUSSIAN_ROUTE: {
                "opt": ["VeryTight"],
                "freq": ["HPModes", "NoRaman"],
            },
            GAUSSIAN_COMMENT: ["comment line 1", "comment line 2"],
        }

        test = FileWriter.write_com(
            geom, theory=theory, outfile=False, **kw_dict
        )

        for line1, line2 in zip(test.splitlines(), ref.splitlines()):
            self.assertEqual(line1.strip(), line2.strip())

    def test_write_inp(self):
        """write orca input file"""
        # like gaussian input files, this compares exact output

        geom = Geometry(self.small_mol)

        ref = """#comment line 1
#comment line 2
! PBE0 Opt Freq def2-SVP CPCM(dichloromethane) D3BJ
%freq
    Temp    298.15
end
%basis
    newGTO            C  "def2-TZVP" end
end
%cpcm
    smd    true
end

*xyz 0 1
C    -1.97696  -2.32718   0.00126
C    -2.36814  -1.29554   0.85518
C    -1.67136  -0.08735   0.85440
C    -0.58210   0.08919   0.00026
C    -0.19077  -0.94241  -0.85309
C    -0.88848  -2.15056  -0.85289
H    -3.22679  -1.43483   1.52790
H    -1.98002   0.72606   1.52699
H     0.66766  -0.80358  -1.52636
H    -0.57992  -2.96360  -1.52585
Cl    0.29699   1.61392  -0.00037
N    -2.73689  -3.64357   0.00188
O    -2.07823  -4.68230   0.00289
O    -3.96579  -3.59263   0.00134
*

"""

        theory = Theory(
            charge=0,
            multiplicity=1,
            method="PBE0",
            basis=BasisSet(
                [Basis("def2-SVP", ["!C"]), Basis("def2-TZVP", ["C"])]
            ),
            empirical_dispersion=EmpiricalDispersion("D3BJ"),
            solvent=ImplicitSolvent("SMD", "dichloromethane"),
            job_type=[FrequencyJob(), OptimizationJob()],
        )

        kw_dict = {ORCA_COMMENT: ["comment line 1", "comment line 2"]}

        test = FileWriter.write_inp(
            geom, theory=theory, outfile=False, **kw_dict
        )

        for line1, line2 in zip(test.splitlines(), ref.splitlines()):
            self.assertEqual(line1.strip(), line2.strip())


        geom = Geometry(self.ts)

        ref = """#test_orbits.out
! M06L printMOs printBasis Opt def2-SVP
%geom
    Constraints
        {B  1  3 C}
        {B  2  5 C}
    end
end

*xyz 0 1
C     1.18854  -0.33581   0.91844
C     0.03710  -0.11108   1.65753
C     1.49009   0.55244  -0.12490
C    -1.46714  -0.56175   0.16760
C    -0.98763   0.39297  -0.71523
O     0.13435   0.25731  -1.31132
H     1.59783  -1.34820   0.87301
H    -0.30053  -0.83534   2.40220
H    -0.29409   0.91445   1.83715
H     2.32894   0.33997  -0.78844
H     1.29000   1.61694   0.01894
H    -1.12283  -1.59079   0.06594
H    -2.43357  -0.40885   0.65412
H    -1.46740   1.39172  -0.71711
*

"""

        theory = Theory(
            charge=0,
            multiplicity=1,
            method="M06-L",
            basis="def2-SVP",
            job_type=[OptimizationJob(constraints={"bonds":[["2", "4"], ["3", "6"]]})],
            comments=["test_orbits.out"],
        )

        kw_dict = {"simple": ["printMOs", "printBasis"]}

        test = FileWriter.write_inp(
            geom, theory=theory, outfile=False, **kw_dict
        )

        for line1, line2 in zip(test.splitlines(), ref.splitlines()):
            self.assertEqual(line1.strip(), line2.strip())


    def test_write_in(self):
        """write psi4 input file"""
        # like gaussian input files, this compares exact output

        geom = Geometry(self.small_mol)

        ref = """#comment line 1
#comment line 2
basis {
    assign    def2-SVP
    assign C  def2-TZVP
}

molecule {
     0 1
     C       -1.97696       -2.32718        0.00126
     C       -2.36814       -1.29554        0.85518
     C       -1.67136       -0.08735        0.85440
     C       -0.58210        0.08919        0.00026
     C       -0.19077       -0.94241       -0.85309
     C       -0.88848       -2.15056       -0.85289
     H       -3.22679       -1.43483        1.52790
     H       -1.98002        0.72606        1.52699
     H        0.66766       -0.80358       -1.52636
     H       -0.57992       -2.96360       -1.52585
    Cl        0.29699        1.61392       -0.00037
     N       -2.73689       -3.64357        0.00188
     O       -2.07823       -4.68230        0.00289
     O       -3.96579       -3.59263        0.00134
}

set {
    T                       298.15
}

nrg = frequencies('PBE0-d3bj')
nrg, wfn = optimize('PBE0-d3bj', return_wfn=True)
"""

        theory = Theory(
            charge=0,
            multiplicity=1,
            method="PBE0",
            basis=BasisSet(
                [Basis("def2-SVP", ["H"]), Basis("def2-TZVP", ["C"])]
            ),
            empirical_dispersion=EmpiricalDispersion("D3BJ"),
            job_type=[FrequencyJob(), OptimizationJob()],
        )

        kw_dict = {
            PSI4_COMMENT: ["comment line 1", "comment line 2"],
            PSI4_JOB: {"optimize": ["return_wfn=True"]},
        }

        test = FileWriter.write_in(
            geom, theory=theory, outfile=False, **kw_dict
        )

        for line1, line2 in zip(test.splitlines(), ref.splitlines()):
            self.assertEqual(line1.strip(), line2.strip())

    def test_write_inq(self):
        """write q-chem input file"""
        # like gaussian input files, this compares exact output

        geom = Geometry(self.small_mol)

        ref = """$rem
    JOB_TYPE             =   OPT
    DFT_D                =   D3_BJ
    BASIS                =   General
    METHOD               =   PBE0
$end

$comments
    comment line 1
    comment line 2
$end

$basis
    H  0
    cc-pVDZ
    ****
    C  0
    aug-cc-pVDZ
    ****
    Cl 0
    aug-cc-pVDZ
    ****
    N  0
    aug-cc-pVDZ
    ****
    O  0
    aug-cc-pVDZ
    ****
$end

$molecule
    0 1
    C     -1.97696     -2.32718      0.00126
    C     -2.36814     -1.29554      0.85518
    C     -1.67136     -0.08735      0.85440
    C     -0.58210      0.08919      0.00026
    C     -0.19077     -0.94241     -0.85309
    C     -0.88848     -2.15056     -0.85289
    H     -3.22679     -1.43483      1.52790
    H     -1.98002      0.72606      1.52699
    H      0.66766     -0.80358     -1.52636
    H     -0.57992     -2.96360     -1.52585
    Cl     0.29699      1.61392     -0.00037
    N     -2.73689     -3.64357      0.00188
    O     -2.07823     -4.68230      0.00289
    O     -3.96579     -3.59263      0.00134
$end
"""

        theory = Theory(
            charge=0,
            multiplicity=1,
            method="PBE0",
            basis=BasisSet(
                [Basis("cc-pVDZ", ["H"]), Basis("aug-cc-pVDZ", ["!H"])]
            ),
            empirical_dispersion=EmpiricalDispersion("D3BJ"),
            job_type=[OptimizationJob()],
        )

        kw_dict = {ORCA_COMMENT: ["comment line 1", "comment line 2"]}

        test = FileWriter.write_inq(
            geom, theory=theory, outfile=False, **kw_dict
        )

        for line1, line2 in zip(test.splitlines(), ref.splitlines()):
            self.assertEqual(line1.strip(), line2.strip())

    def test_read_crest(self):
        fname = os.path.join(prefix, "test_files/good.crest")
        res = FileReader(fname, conf_name=False)
        self.assertDictEqual(
            res.other,
            {
                "avg_energy": 0.00027409943594479446,
                "temperature": 298.15,
                "energy": -79.18948,
                "entropy": 0.010994255863855447,
                "free_energy": -0.003278038021735129,
                "best_pop": 70.368,
                "finished": True,
                "error": None,
            },
        )


if __name__ == "__main__":
    unittest.main()



import time
import unittest
from os.path import dirname

import numpy as np
from AaronTools.geometry import Geometry

prefix = dirname(__file__)


def rmsd_tol(geom, superTight=False, superLoose=False):
    """
    Automatically determine a reasonable rmsd tolerance for the input
    geometry based on its size and number of atoms
    """
    tolerance = len(geom.atoms) ** (
        2 - int(superTight) + int(superLoose)
    ) * np.sqrt(np.finfo(float).eps)

    com = geom.COM()
    max_d = None
    for atom in geom.atoms:
        d = np.linalg.norm(atom.coords - com)
        if max_d is None or d > max_d:
            max_d = d

    tolerance *= max_d * (2 - int(superTight) + int(superLoose))
    tolerance = tolerance ** (2 / (4 - int(superTight) + int(superLoose)))
    return tolerance


def check_atom_list(ref, comp):
    rv = True
    for i, j in zip(ref, comp):
        rv &= i.__repr__() == j.__repr__()
    return rv


def validate(test, ref, thresh=None, heavy_only=False, sort=True, debug=False):
    """
    Validates `test` geometry against `ref` geometry
    Returns: True if validation passed, False if failed

    :test: the geometry to validate
    :ref: the reference geometry
    :thresh: the RMSD threshold
        if thresh is a number: use that as threshold
        if thresh is None: use rmsd_tol() to determine
        if thresh is "tight": use rmsd_tol(superTight=True)
        if thresh is "loose": use rmsd_tol(superLoose=True)
    :sort: allow canonical sorting of atoms
    :debug: print info useful for debugging
    """
    if debug:
        print("before alignment")
        print(ref.write("ref", outfile=False))
        print(test.write("test", outfile=False))

    if thresh is None:
        thresh = rmsd_tol(ref)
    try:
        thresh = float(thresh)
    except ValueError:
        if thresh.lower() == "tight":
            thresh = rmsd_tol(ref, superTight=True)
        elif thresh.lower() == "loose":
            thresh = rmsd_tol(ref, superLoose=True)
        else:
            raise ValueError("Bad threshold provided")

    # elements should all be the same
    t_el = sorted([t.element for t in test.atoms])
    r_el = sorted([r.element for r in ref.atoms])
    if len(t_el) != len(r_el):
        if debug:
            print(
                "wrong number of atoms: {} (test) vs. {} (ref)".format(
                    len(t_el), len(r_el)
                )
            )
        return False

    for t, r in zip(t_el, r_el):
        if t != r:
            if debug:
                print("elements don't match")
            return False
    # and RMSD should be below a threshold
    rmsd = test.RMSD(
        ref, align=debug, heavy_only=heavy_only, sort=sort, debug=debug
    )
    if debug:
        print("after alignment")
        print(ref.write("ref", outfile=False))
        print(test.write("test", outfile=False))

    if debug:
        print("RMSD:", rmsd[2], "\tTHRESH:", thresh)
        rmsd[0].write("ref")
        rmsd[1].write("test")
        rmsd = rmsd[2]

    return rmsd < thresh


class TestWithTimer(unittest.TestCase):
    test_count = 0
    total_time = 0
    this_class = None
    last_class = None
    last_result = None
    errors = 0
    fails = 0
    last_errors = 0
    last_fails = 0

    @classmethod
    def setUpClass(cls):
        TestWithTimer.total_time = time.time()

    @classmethod
    def tearDownClass(cls):
        TestWithTimer.total_time = time.time() - TestWithTimer.total_time
        print(TestWithTimer.get_status())

        if TestWithTimer.errors - TestWithTimer.last_errors:
            status = "ERROR"
        elif TestWithTimer.fails - TestWithTimer.last_fails:
            status = "FAIL"
        else:
            status = "ok"
        TestWithTimer.last_errors = TestWithTimer.errors
        TestWithTimer.last_fails = TestWithTimer.fails

        print(
            "Ran %d test in %.4fs  %s"
            % (
                TestWithTimer.last_result.testsRun - TestWithTimer.test_count,
                TestWithTimer.total_time,
                status,
            )
        )
        TestWithTimer.test_count = TestWithTimer.last_result.testsRun
        print(unittest.TextTestResult.separator2)

    @classmethod
    def get_status(cls):
        if TestWithTimer.errors != len(TestWithTimer.last_result.errors):
            status = "ERROR"
        elif TestWithTimer.fails != len(TestWithTimer.last_result.failures):
            status = "FAIL"
        else:
            status = "ok"
        TestWithTimer.errors = len(TestWithTimer.last_result.errors)
        TestWithTimer.fails = len(TestWithTimer.last_result.failures)
        return status

    def setUp(self):
        self.start_time = time.time()

    def tearDown(self):
        t = time.time() - self.start_time
        TestWithTimer.last_result = self._outcome.result
        TestWithTimer.this_class, self.test_name = self.id().split(".")[-2:]

        status = TestWithTimer.get_status()
        if TestWithTimer.this_class != TestWithTimer.last_class:
            TestWithTimer.last_class = TestWithTimer.this_class
            print(TestWithTimer.this_class)
        else:
            print(status)

        print(
            "\b   %2d. %-30s %.4fs  "
            % (
                TestWithTimer.last_result.testsRun - TestWithTimer.test_count,
                self.test_name,
                t,
            ),
            end="",
        )



#!/usr/bin/env python3

from AaronTools.config import Config


def main(args):
    config = Config(args.config)
    default_keys = tuple("/" + k for k in config["DEFAULT"])
    print("Configuration:")
    for key, val in config.as_dict().items():
        if key.endswith(default_keys):
            continue
        if key.startswith(("Substitution", "Mapping")):
            print("  ", key, ": ", "\n\t".join(val.split("\n")), sep="")

    template = "  {:20s} {:20s} {:s}"
    print()
    print("Structures to run:")
    print(template.format("Name", "Type", "Change"))
    for key, val in config._changes.items():
        print(
            template.format(
                key if key else "original", str(val[1]), str(val[0])
            )
        )


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description="Lists the substitutions and mappings parsed from an AaronJr configuration file"
    )
    parser.add_argument("config", help="The AaronJr configuration file")
    args = parser.parse_args()
    main(args)



#!/usr/bin/env python3

import os
import argparse
import sys

from AaronTools.geometry import Geometry
from AaronTools.component import Component

libaddlig_parser = argparse.ArgumentParser(
    description='add a ligand to your personal library',
    formatter_class=argparse.RawTextHelpFormatter
)

libaddlig_parser.add_argument(
    'infile', metavar='input file',
    type=str,
    default=None,
    help='a coordinate file'
)

libaddlig_parser.add_argument(
    '-n', '--name',
    type=str,
    required=False,
    default=None,
    dest='name',
    help="Name of ligand being added to the library\n" +
    "if no name is given, the ligand will be printed to STDOUT"
)

args = libaddlig_parser.parse_args()

cat = Geometry(args.infile)

if cat.center is None:
    cat.detect_components()

#move center to origin
cat.coord_shift(-cat.COM(targets=cat.center))

ligands = cat.components
lig_index = 0

#if there's multiple ligands, ask which one we should add
if len(ligands) > 1:
    print("multiple ligands found:")
    for i, ligand in enumerate(ligands):
        print("ligand %i:" % i)
        print(ligand.write(outfile=False))

    lig_index = int(input("What is the number of the ligand you would like to add?\n"))

new_ligand = ligands[lig_index]
new_ligand.comment = "K:" + ",".join(
    [str(new_ligand.atoms.index(key) + 1) for key in new_ligand.key_atoms]
) + ';'

if args.name is None:
    print(new_ligand.write(outfile=False))
else:
    lig_file = os.path.join(os.path.dirname(Component.AARON_LIBS), args.name + '.xyz')
    if os.path.exists(lig_file):
        overwrite = input(
            "%s already exists.\nWould you like to overwrite it? (YES/no)\n" % lig_file
        )
        if overwrite != "YES":
            print("%s to overwrite, not overwriting" % overwrite)
            sys.exit(0)

    new_ligand.write(outfile=lig_file)



"""
used for specifying integration grids for DFT calculations
"""

import re
from AaronTools import addlogger
from AaronTools.theory import (
    GAUSSIAN_ROUTE, PSI4_SETTINGS, ORCA_ROUTE, ORCA_BLOCKS, QCHEM_REM
)


class IntegrationGrid:
    """
    used to try to keep integration grid settings more
    easily when writing different input files
    """
    
    LOG = None
    
    def __init__(self, name):
        """
        name: str, Gaussian keyword (e.g. SuperFineGrid),
                   ORCA keyword (e.g. Grid7),
                   or "(radial, angular)"
        ORCA can only use ORCA grid keywords
        Gaussian can use its keywords and will try to use ORCA keywords, and
        can use "(radial, angular)" or radialangular
        Psi4 will use "(radial, angular)" and will try to use ORCA or Gaussian keywords
        """
        self.name = name

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        if self.name == other.name:
            return True
        return self.get_gaussian()[0] == other.get_gaussian()[0]

    def copy(self):
        new_dict = dict()
        for key, value in self.__dict__.items():
            try:
                new_dict[key] = value.copy()
            except AttributeError:
                new_dict[key] = value
                # ignore chimerax objects so seqcrow doesn't print a
                # warning when a geometry is copied
                if "chimerax" in value.__class__.__module__:
                    continue
                if value.__class__.__module__ != "builtins":
                    self.LOG.warning(
                        "No copy method for {}: in-place changes may occur".format(
                            type(value)
                        )
                    )
        
        return self.__class__(**new_dict)
 
    def get_gaussian(self):
        """
        gets gaussian integration grid info and a warning as tuple(dict, str or None)
        dict is of the form {GAUSSIAN_ROUTE:[x]}
        """
        if self.name.lower() == "ultrafine":
            return ({GAUSSIAN_ROUTE:{"Integral":["grid=UltraFine"]}}, None)

        elif self.name.lower() == "finegrid":
            return ({GAUSSIAN_ROUTE:{"Integral":["grid=FineGrid"]}}, None)

        elif self.name.lower() == "superfinegrid":
            return ({GAUSSIAN_ROUTE:{"Integral":["grid=SuperFineGrid"]}}, None)

        # Grids available in ORCA but not Gaussian
        # uses n_rad from K-Kr as specified in ORCA 4.2.1 manual (section 9.3)
        elif self.name.lower() == "grid2":
            out_dict = {
                GAUSSIAN_ROUTE: {
                    "Integral":["grid=45110"]
                }
            }
            return (out_dict, "Approximating ORCA Grid 2")

        elif self.name.lower() == "grid3":
            out_dict = {
                GAUSSIAN_ROUTE: {
                    "Integral":["grid=45194"]
                }
            }
            return (out_dict, "Approximating ORCA Grid 3")

        elif self.name.lower() == "grid4":
            out_dict = {
                GAUSSIAN_ROUTE: {
                    "Integral":["grid=45302"]
                }
            }
            return (out_dict, "Approximating ORCA Grid 4")

        elif self.name.lower() == "grid5":
            out_dict = {
                GAUSSIAN_ROUTE: {
                    "Integral":["grid=50434"]
                }
            }
            return (out_dict, "Approximating ORCA Grid 5")

        elif self.name.lower() == "grid6":
            out_dict = {
                GAUSSIAN_ROUTE: {
                    "Integral":["grid=55590"]
                }
            }
            return (out_dict, "Approximating ORCA Grid 6")

        elif self.name.lower() == "grid7":
            out_dict = {
                GAUSSIAN_ROUTE: {
                    "Integral":["grid=60770"]
                }
            }
            return (out_dict, "Approximating ORCA Grid 7")

        # grid format may be (int, int)
        # or just int
        match = re.match(r"\(\s*?(\d+)\s*?,\s*?(\d+)?\s*\)", self.name)
        match_digit = re.match(r"-?\d+?", self.name)
        if match:
            r_pts = int(match.group(1))
            a_pts = int(match.group(2))
            out_dict = {
                GAUSSIAN_ROUTE: {
                    "Integral":["grid=%i%03i" % (r_pts, a_pts)]
                }
            }
            return (out_dict, None)

        elif match_digit:
            out_dict = {
                GAUSSIAN_ROUTE: {
                    "Integral":["grid=%s" % self.name]
                }
            }
            return (out_dict, None)

        out_dict = {
            GAUSSIAN_ROUTE: {
                "Integral":["grid=%s" % self.name]
            }
        }
        return (out_dict, "grid may not be available in Gaussian")

    def get_orca(self):
        """
        translates grid to something ORCA accepts
        returns tuple(dict(ORCA_ROUTE:[self.name]), None) or
        tuple(dict(ORCA_BLOCKS:{"method":[str]}), None)
        """
        if self.name.lower() == "ultrafine":
            out_dict = {
                ORCA_BLOCKS: {
                    "method": [
                        "AngularGrid     Lebedev590",
                        "IntAcc          4.0",
                    ]
                }
            }
            return (out_dict, "approximating UltraFineGrid")

        elif self.name.lower() == "finegrid":
            out_dict = {
                ORCA_BLOCKS: {
                    "method": [
                        "AngularGrid     Lebedev302",
                        "IntAcc          4.0",
                    ]
                }
            }

            return (out_dict, "approximating FineGrid")

        elif self.name.lower() == "superfinegrid":
            #radial is 175 for 1st row, 250 for later rows
            out_dict = {
                ORCA_ROUTE: [
                    "Grid7",
                    "FinalGrid7",
                ],
            }

            return (
                out_dict,
                "could not set SuperFineGrid equivalent - using largest ORCA grid keyword",
            )

        elif re.match("defgrid\d", self.name.lower()):
            warnings = None
            grid_n = re.match("defgrid(\d)", self.name.lower()).group(1)
            if int(grid_n) < 1 or int(grid_n) > 3:
                warnings = "grid may not be available"
            out_dict = {
                ORCA_ROUTE: [self.name]
            }
            return (out_dict, warnings)

        elif self.name.lower().startswith("grid"):
            # orca grid keyword
            out_dict = {
                ORCA_ROUTE: [
                    self.name,
                    "Final%s" % self.name,
                ]
            }
            return (out_dict, None)

        # grid format may be (int, int)
        match = re.match(r"\(\s*?(\d+)\s*?,\s*?(\d+)?\s*\)", self.name)
        if match:
            r_pts = int(match.group(1))
            a_pts = int(match.group(2))
            int_acc = -((r_pts / -5) + 2 - 8) / 3
            out_dict = {
                ORCA_BLOCKS: {
                    "method": [
                        "AngularGrid     Lebedev%i" % a_pts,
                        "IntAcc          %.1f" % int_acc,
                    ]
                }
            }
            return (out_dict, None)

        raise RuntimeError(
            "could not determine acceptable ORCA grid settings for %s" % self.name
        )

    def get_psi4(self):
        """
        returns ({PSI4_SETTINGS:{"dft_radial_points":["n"], "dft_spherical_points":["m"]}}, warning)
        """
        if self.name.lower() == "ultrafine":
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points": ["99"],
                    "dft_spherical_points": ["590"],
                }
            }
            return (out_dict, None)

        elif self.name.lower() == "finegrid":
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points": ["75"],
                    "dft_spherical_points": ["302"],
                }
            }
            return (out_dict, None)

        elif self.name.lower() == "superfinegrid":
            # radial is 175 for 1st row, 250 for later rows
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points": ["250"],
                    "dft_spherical_points": ["974"],
                }
            }
            return (out_dict, "Approximating Gaussian SuperFineGrid")

        # uses radial from K-Kr as specified in ORCA 4.2.1 manual (section 9.3)
        elif self.name.lower() == "grid2":
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points":["45"],
                    "dft_spherical_points": ["110"],
                }
            }
            return (out_dict, "Approximating ORCA Grid 2")

        elif self.name.lower() == "grid3":
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points": ["45"],
                    "dft_spherical_points":["194"],
                }
            }
            return (out_dict, "Approximating ORCA Grid 3")

        elif self.name.lower() == "grid4":
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points":["45"],
                    "dft_spherical_points": ["302"],
                }
            }
            return (out_dict, "Approximating ORCA Grid 4")
 
        elif self.name.lower() == "grid5":
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points":["50"],
                    "dft_spherical_points": ["434"],
                }
            }
            return (out_dict, "Approximating ORCA Grid 5")

        elif self.name.lower() == "grid6":
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points": ["55"],
                    "dft_spherical_points": ["590"],
                }
            }
            return (out_dict, "Approximating ORCA Grid 6")

        elif self.name.lower() == "grid7":
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points": ["60"],
                    "dft_spherical_points": ["770"],
                }
            }
            return (out_dict, "Approximating ORCA Grid 7")

        # grid format may be (int, int)
        match = re.match(r"\(\s*?(\d+)\s*?,\s*?(\d+)?\s*\)", self.name)
        if match:
            r_pts = int(match.group(1))
            a_pts = int(match.group(2))
            out_dict = {
                PSI4_SETTINGS: {
                    "dft_radial_points": [r_pts],
                    "dft_spherical_points": [a_pts],
                }
            }
            return (out_dict, None)

        raise RuntimeError(
            "could not determine acceptable Psi4 grid settings for %s" % self.name
        )

    def get_qchem(self):
        if self.name.lower() == "sg-1":
            out_dict = {
                QCHEM_REM: {
                    "XC_GRID": "1",
                }
            }
            return (out_dict, None)
        
        elif self.name.lower() == "sg-2":
            out_dict = {
                QCHEM_REM: {
                    "XC_GRID": "2",
                }
            }
            return (out_dict, None)
        
        elif self.name.lower() == "sg-3":
            out_dict = {
                QCHEM_REM: {
                    "XC_GRID": "3",
                }
            }
            return (out_dict, None)
        
        elif self.name.lower() == "ultrafine":
            out_dict = {
                QCHEM_REM: {
                    "XC_GRID": "%2i%06i" % (99, 590),
                }
            }
            return (out_dict, None)

        elif self.name.lower() == "finegrid":
            out_dict = {
                QCHEM_REM: {
                    "XC_GRID": "%2i%06i" % (75, 302),
                }
            }
            return (out_dict, None)

        elif self.name.lower() == "superfinegrid":
            # radial is 175 for 1st row, 250 for later rows
            out_dict = {
                QCHEM_REM: {
                    "XC_GRID": "%2i%06i" % (250, 974),
                }
            }
            return (out_dict, "Approximating Gaussian SuperFineGrid")

        # grid format may be "(int, int)"
        match = re.match(r"\(\s*?(\d+)\s*?,\s*?(\d+)?\s*\)", self.name)
        if match:
            r_pts = int(match.group(1))
            a_pts = int(match.group(2))
            out_dict = {
                QCHEM_REM: {
                    "XC_GRID": "%2i%06i" % (r_pts, a_pts),
                }
            }
            return (out_dict, None)

        raise RuntimeError(
            "could not determine acceptable QChem grid settings for %s" % self.name
        )



#! /usr/bin/env python3
import os
import unittest
from copy import copy

from AaronTools.comp_output import CompOutput
from AaronTools.test import TestWithTimer, prefix


class TestCompOutput(TestWithTimer):
    # with frequencies
    normal = CompOutput(os.path.join(prefix, "test_files", "normal.log"))
    died = CompOutput(os.path.join(prefix, "test_files", "died.log"))
    error = CompOutput(os.path.join(prefix, "test_files", "error.log"))
    # optimization
    opt_run = CompOutput(os.path.join(prefix, "test_files", "opt_running.log"))
    opt_norm = CompOutput(os.path.join(prefix, "test_files", "opt_normal.log"))

    def test_get_progress(self):
        test = [TestCompOutput.died.get_progress()]
        test += [TestCompOutput.error.get_progress()]
        test += [TestCompOutput.normal.get_progress()]
        test += [TestCompOutput.opt_run.get_progress()]

        ref = ["Progress not found"]
        ref += [
            "Max Force:0.003790/NO  RMS Force:0.000887/NO   Max Disp:1.095802/NO   RMS Disp:0.158612/NO"
        ]
        ref += [
            "Max Force:0.000016/YES RMS Force:0.000006/YES  Max Disp:0.011148/NO   RMS Disp:0.003723/NO"
        ]
        ref += [
            "Max Force:3.606006/NO  RMS Force:0.254588/NO   Max Disp:1.656082/NO   RMS Disp:0.279091/NO"
        ]

        for t, r in zip(test, ref):
            self.assertTrue(t == r)

    def test_grab_thermo(self):
        logs = [
            TestCompOutput.normal,
            TestCompOutput.error,
            TestCompOutput.opt_norm,
        ]

        tmp = logs[0]
        test = [
            [
                copy(tmp.energy),
                copy(tmp.enthalpy),
                copy(tmp.free_energy),
                copy(tmp.calc_Grimme_G()),
                copy(tmp.calc_G_corr(method="QHARM")),
            ]
        ]
        ref = [
            [
                -1856.01865834,
                -1855.440611,
                -1855.538011,
                -1855.5328046892625,
                0.486148,
            ]
        ]

        tmp = logs[1]
        test += [
            [copy(tmp.energy), copy(tmp.enthalpy), copy(tmp.free_energy), None]
        ]
        ref += [[1.00054354503, None, None, None]]

        tmp = logs[2]
        test += [
            [copy(tmp.energy), copy(tmp.enthalpy), copy(tmp.free_energy), None]
        ]
        ref += [[-3511.5160178, None, None, None]]

        for i, r in enumerate(ref):
            if r[3] is None:
                self.assertRaises(AttributeError, logs[i].calc_Grimme_G)
            for j in range(3):
                if r[j] is None:
                    self.assertTrue(test[i][j] is None)
                else:
                    self.assertTrue(test[i][j] == r[j])

    def test_crest(self):
        test = CompOutput(
            os.path.join(prefix, "test_files", "good.crest"), conf_name=False
        )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
import os
import unittest
import numpy as np

from AaronTools.component import Component
from AaronTools.geometry import Geometry
from AaronTools.substituent import Substituent
from AaronTools.test import TestWithTimer, prefix, validate


class TestComponent(TestWithTimer):
    # simple geometries
    benz = Component(os.path.join(prefix, "test_files", "benzene.xyz"))
    benz_Cl = Component(os.path.join(prefix, "test_files", "benzene_4-Cl.xyz"))
    benz_NO2_Cl = Component(
        os.path.join(prefix, "test_files", "benzene_1-NO2_4-Cl.xyz")
    )
    benz_OH_Cl = Component(
        os.path.join(prefix, "test_files", "benzene_1-OH_4-Cl.xyz")
    )
    benz_Ph_Cl = Component(
        os.path.join(prefix, "test_files", "benzene_1-Ph_4-Cl.xyz")
    )
    Et_NO2 = Component(os.path.join(prefix, "test_files", "Et_1-NO2.xyz"))
    pent = Component(os.path.join(prefix, "test_files", "pentane.xyz"))

    # ligands
    RQ_tBu = Component(os.path.join(prefix, "test_files", "R-Quinox-tBu3.xyz"))
    for a in RQ_tBu.find("P"):
        a.add_tag("key")
    tri = Component(
        os.path.join(prefix, "test_files", "ligands", "squaramide.xyz")
    )

    def is_member(self, valid, test):
        for a in valid:
            try:
                test.remove(a)
            except KeyError:
                return False
        if len(test) == 0:
            return True
        else:
            return False

    def test_substitute(self):
        mol = TestComponent.benz.copy()
        benz_Cl = TestComponent.benz_Cl.copy()
        benz_NO2_Cl = TestComponent.benz_NO2_Cl.copy()
        benz_OH_Cl = TestComponent.benz_OH_Cl.copy()
        benz_Ph_Cl = TestComponent.benz_Ph_Cl.copy()

        mol.substitute(Substituent("Cl"), "11")
        res = validate(mol, benz_Cl)
        self.assertTrue(res)

        mol.substitute(Substituent("NO2"), "12", "1")
        res = validate(mol, benz_NO2_Cl, sort=True)
        self.assertTrue(res)

        mol.substitute(Substituent("OH"), "NO2")
        res = validate(mol, benz_OH_Cl, sort=True)
        self.assertTrue(res)

        mol.substitute(Substituent("Ph"), ["12", "12.*"])
        res = validate(mol, benz_Ph_Cl, sort=True, thresh="loose")
        self.assertTrue(res)

    def test_detect_backbone(self):
        geom = TestComponent.RQ_tBu.copy()

        backbone = geom.find("1,2,7-20")
        Me = geom.find("3,21-23")
        tBu = geom.find("4-6,24-59")

        try:
            test_Me = set(geom.find(["Me", "CH3"]))
            test_tBu = set(geom.find("tBu"))
            test_backbone = set(geom.find("backbone"))
        except LookupError:
            self.assertTrue(False)

        self.assertTrue(self.is_member(Me, test_Me))
        self.assertTrue(self.is_member(tBu, test_tBu))
        self.assertTrue(self.is_member(backbone, test_backbone))

    def test_minimize_torsion(self):
        ref = Component(
            os.path.join(prefix, "ref_files", "minimize_torsion.xyz")
        )

        geom = TestComponent.benz.copy()
        geom.substitute(Substituent("tBu"), "12")
        geom.substitute(Substituent("Ph"), "10")
        geom.substitute(Substituent("OH"), "7")
        geom.minimize_sub_torsion(allow_planar=True)
        self.assertTrue(validate(geom, ref, heavy_only=True))

    def test_sub_rotate(self):
        geom = TestComponent.RQ_tBu.copy()
        geom.sub_rotate("4", angle=60)
        geom.sub_rotate("6", angle=60)
        geom.sub_rotate("5", angle=60)
        ref = Component(os.path.join(prefix, "ref_files", "sub_rotate.xyz"))
        self.assertTrue(validate(geom, ref, heavy_only=True))


ONLYSOME = False


def suite():
    suite = unittest.TestSuite()
    # suite.addTest(TestComponent("test_detect_backbone"))
    # suite.addTest(TestComponent("test_sub_rotate"))
    # suite.addTest(TestComponent("test_substitute"))
    suite.addTest(TestComponent("test_minimize_torsion"))
    # suite.addTest(TestComponent("test_cone_angle"))
    return suite


if __name__ == "__main__" and ONLYSOME:
    runner = unittest.TextTestRunner()
    runner.run(suite())
elif __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

"""Generates prime numbers"""
import sys
from math import sqrt
from os import makedirs, path, access, W_OK

from AaronTools.const import AARONLIB


class Primes:
    """
    find and cache prime numbers
    """
    primes = [2, 3]
    clean = False
    cache = path.join(AARONLIB, "cache", "primes.dat")

    def __init__(self, clean=False, cache=None):
        Primes.clean = clean
        if cache is not None:
            Primes.cache = cache
        if Primes.clean or (not path.exists(Primes.cache) and access(Primes.cache, W_OK)):
            prime_dir, _ = path.split(Primes.cache)
            if not path.exists(prime_dir):
                makedirs(prime_dir)

            with open(Primes.cache, "w") as f:
                f.writelines([str(i) + "\n" for i in Primes.primes])
            f.close()

    @classmethod
    def next_prime(cls):
        """determine the next prime number"""
        # first return the ones we already found
        if path.exists(Primes.cache):
            with open(Primes.cache) as f:
                for line in f:
                    prime = int(line.strip())
                    if prime in cls.primes:
                        continue
                    cls.primes += [prime]
                    yield prime
            f.close()

        # then start generating new ones
        test_prime = cls.primes[-1] + 2
        while True:
            max_check = sqrt(test_prime)
            for prime in cls.primes[1:]:
                if test_prime % prime == 0:
                    test_prime += 2
                    break
                if prime > max_check:
                    cls.primes += [test_prime]
                    cls.store_prime(test_prime)
                    yield test_prime
                    test_prime += 2
                    break
            else:
                cls.primes += [test_prime]
                cls.store_prime(test_prime)
                yield test_prime
                test_prime += 2

    @classmethod
    def store_prime(cls, prime):
        """add the prime number to the cache"""
        if not path.exists(path.dirname(cls.cache)):
            return
        with open(cls.cache, "a") as f:
            if hasattr(prime, "__iter__"):
                f.writelines([str(i) + "\n" for i in prime])
            else:
                f.write(str(prime) + "\n")
        f.close()

    @classmethod
    def list(cls, n):
        """list the first n prime numbers"""
        rv = [prime for i, prime in enumerate(cls.primes) if i < n]
        new = False
        if len(cls.primes) < n:
            if path.exists(cls.cache):
                cur_primes = len(cls.primes)
                k = 0
                with open(cls.cache) as f:
                    for line in f:
                        prime = int(line)
                        k += 1
                        if new or prime not in cls.primes:
                            new = True
                            cls.primes += [prime]
                            rv += [prime]
                            if len(cls.primes) >= n:
                                break
            
            new_primes = []
        
            test_prime = cls.primes[-1] + 2
            while len(cls.primes) < n:
                # this is similar to next_prime, though it doesn't
                # cache each new prime number as it finds them
                # it saves them as a group
                # writing to a file is slow as heck
                max_check = sqrt(test_prime)
                for prime in cls.primes[1:]:
                    if test_prime % prime == 0:
                        test_prime += 2
                        break
                    if prime > max_check:
                        cls.primes += [test_prime]
                        new_primes += [test_prime]
                        test_prime += 2
                        break
                else:
                    cls.primes += [test_prime]
                    new_primes += [test_prime]
                    test_prime += 2

            rv += new_primes
            # print(new_primes)
            cls.store_prime(new_primes)
            
        return rv

    @classmethod
    def primes_below(cls, le):
        """list the primes that are less than or equal to le"""
        rv = [p for p in cls.primes if p <= le]
        if cls.primes[-1] <= le:
            for i, prime in enumerate(cls.next_prime()):
                if prime >= le:
                    break
                rv += [prime]
        return rv

if __name__ == "__main__":
    Primes(clean=True)
    print(Primes.list(int(sys.argv[1])))
